# encoding: utf-8
"""
ospfroute.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.protocol.ip import IP
from exabgp.bgp.message.notification import Notify
from exabgp.util import ordinal

#     https://tools.ietf.org/html/rfc7752#section-3.2.3

#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |  Route Type   |
#     +-+-+-+-+-+-+-+-+

#  Route Type can be:
#    Intra-Area (0x1)
#    Inter-Area (0x2)
#    External 1 (0x3)
#    External 2 (0x4)
#    NSSA 1 (0x5)
#    NSSA 2 (0x6)

# ================================================================== OSPF_ROUTE_TYPE


OSPF_ROUTE = {
	1:	'intra-area',
	2:	'inter-area',
	3:	'external-1',
	4:	'external-2',
	5:	'nssa-1',
	6:	'nssa-2'
}


class OspfRoute(object):
	def __init__ (self, ospf_type, packed=None):
		self.ospf_type = ospf_type
		self._packed = packed

	@classmethod
	def unpack (cls, data):
		if len(data) == 1:
			ospf_type = unpack('!B',data[0:1])[0]
		return cls(ospf_type=ospf_type)

	def json (self):
		content = '"ospf-route-type": %s' % self.ospf_type
		return content

	def __eq__ (self, other):
		return self.ospf_type == other.ospf_type

	def __neq__ (self, other):
		return self.ospf_type != other.ospf_type

	def __lt__ (self, other):
		raise RuntimeError('Not implemented')

	def __le__ (self, other):
		raise RuntimeError('Not implemented')

	def __gt__ (self, other):
		raise RuntimeError('Not implemented')

	def __ge__ (self, other):
		raise RuntimeError('Not implemented')

	def __str__ (self):
		return ':'.join('%02X' % ordinal(_) for _ in self._packed)

	def __repr__ (self):
		return self.__str__()

	def __hash__ (self):
		return hash(str(self))

	def pack (self):
		return self._packed



# encoding: utf-8
"""
version.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character

# =================================================================== Version
#


class Version (int):
	def pack (self):
		return character(self)



# encoding: utf-8
"""
icmp.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.resource import Resource


# ============================================================== ICMP Code Field
# https://www.iana.org/assignments/icmp-parameters

class ICMPType (Resource):
	NAME = 'icmp type'

	ECHO_REPLY               = 0x00
	# DESTINATION_UNREACHEABLE = 0x03
	UNREACHABLE              = 0x03
	SOURCE_QUENCH            = 0x04
	REDIRECT                 = 0x05
	ECHO_REQUEST             = 0x08
	ROUTER_ADVERTISEMENT     = 0x09
	ROUTER_SOLICIT           = 0x0A
	TIME_EXCEEDED            = 0x0B
	PARAMETER_PROBLEM        = 0x0C
	TIMESTAMP                = 0x0D
	TIMESTAMP_REPLY          = 0x0E
	INFO_REQUEST             = 0x0F
	INFO_REPLY               = 0x10
	MASK_REQUEST             = 0x11
	MASK_REPLY               = 0x12
	TRACEROUTE               = 0x1E

	codes = dict ((k.lower().replace('_','-'),v) for (k,v) in {
		'ECHO_REPLY':               ECHO_REPLY,
		'UNREACHABLE':              UNREACHABLE,
		'SOURCE_QUENCH':            SOURCE_QUENCH,
		'REDIRECT':                 REDIRECT,
		'ECHO_REQUEST':             ECHO_REQUEST,
		'ROUTER_ADVERTISEMENT':     ROUTER_ADVERTISEMENT,
		'ROUTER_SOLICIT':           ROUTER_SOLICIT,
		'TIME_EXCEEDED':            TIME_EXCEEDED,
		'PARAMETER_PROBLEM':        PARAMETER_PROBLEM,
		'TIMESTAMP':                TIMESTAMP,
		'TIMESTAMP_REPLY':          TIMESTAMP_REPLY,
		'INFO_REQUEST':             INFO_REQUEST,
		'INFO_REPLY':               INFO_REPLY,
		'MASK_REQUEST':             MASK_REQUEST,
		'MASK_REPLY':               MASK_REPLY,
		'TRACEROUTE':               TRACEROUTE,
	}.items())

	names = dict([(r,l) for (l,r) in codes.items()])


# https://www.iana.org/assignments/icmp-parameters
class ICMPCode (Resource):
	NAME = 'icmp code'

	# Destination Unreacheable (type 3)
	NETWORK_UNREACHABLE                   = 0x0
	HOST_UNREACHABLE                      = 0x1
	PROTOCOL_UNREACHABLE                  = 0x2
	PORT_UNREACHABLE                      = 0x3
	FRAGMENTATION_NEEDED                  = 0x4
	SOURCE_ROUTE_FAILED                   = 0x5
	DESTINATION_NETWORK_UNKNOWN           = 0x6
	DESTINATION_HOST_UNKNOWN              = 0x7
	SOURCE_HOST_ISOLATED                  = 0x8
	DESTINATION_NETWORK_PROHIBITED        = 0x9
	DESTINATION_HOST_PROHIBITED           = 0xA
	NETWORK_UNREACHABLE_FOR_TOS           = 0xB
	HOST_UNREACHABLE_FOR_TOS              = 0xC
	COMMUNICATION_PROHIBITED_BY_FILTERING = 0xD
	HOST_PRECEDENCE_VIOLATION             = 0xE
	PRECEDENCE_CUTOFF_IN_EFFECT           = 0xF

	# Redirect (Type 5)
	REDIRECT_FOR_NETWORK                  = 0x0
	REDIRECT_FOR_HOST                     = 0x1
	REDIRECT_FOR_TOS_AND_NET              = 0x2
	REDIRECT_FOR_TOS_AND_HOST             = 0x3

	# Time Exceeded (Type 11)
	TTL_EQ_ZERO_DURING_TRANSIT            = 0x0
	TTL_EQ_ZERO_DURING_REASSEMBLY         = 0x1

	# parameter Problem (Type 12)
	REQUIRED_OPTION_MISSING               = 0x1
	IP_HEADER_BAD                         = 0x2

	codes = dict ((k.lower().replace('_','-'),v) for (k,v) in {
		'NETWORK_UNREACHABLE':                   NETWORK_UNREACHABLE,
		'HOST_UNREACHABLE':                      HOST_UNREACHABLE,
		'PROTOCOL_UNREACHABLE':                  PROTOCOL_UNREACHABLE,
		'PORT_UNREACHABLE':                      PORT_UNREACHABLE,
		'FRAGMENTATION_NEEDED':                  FRAGMENTATION_NEEDED,
		'SOURCE_ROUTE_FAILED':                   SOURCE_ROUTE_FAILED,
		'DESTINATION_NETWORK_UNKNOWN':           DESTINATION_NETWORK_UNKNOWN,
		'DESTINATION_HOST_UNKNOWN':              DESTINATION_HOST_UNKNOWN,
		'SOURCE_HOST_ISOLATED':                  SOURCE_HOST_ISOLATED,
		'DESTINATION_NETWORK_PROHIBITED':        DESTINATION_NETWORK_PROHIBITED,
		'DESTINATION_HOST_PROHIBITED':           DESTINATION_HOST_PROHIBITED,
		'NETWORK_UNREACHABLE_FOR_TOS':           NETWORK_UNREACHABLE_FOR_TOS,
		'HOST_UNREACHABLE_FOR_TOS':              HOST_UNREACHABLE_FOR_TOS,
		'COMMUNICATION_PROHIBITED_BY_FILTERING': COMMUNICATION_PROHIBITED_BY_FILTERING,
		'HOST_PRECEDENCE_VIOLATION':             HOST_PRECEDENCE_VIOLATION,
		'PRECEDENCE_CUTOFF_IN_EFFECT':           PRECEDENCE_CUTOFF_IN_EFFECT,
		'REDIRECT_FOR_NETWORK':                  REDIRECT_FOR_NETWORK,
		'REDIRECT_FOR_HOST':                     REDIRECT_FOR_HOST,
		'REDIRECT_FOR_TOS_AND_NET':              REDIRECT_FOR_TOS_AND_NET,
		'REDIRECT_FOR_TOS_AND_HOST':             REDIRECT_FOR_TOS_AND_HOST,
		'TTL_EQ_ZERO_DURING_TRANSIT':            TTL_EQ_ZERO_DURING_TRANSIT,
		'TTL_EQ_ZERO_DURING_REASSEMBLY':         TTL_EQ_ZERO_DURING_REASSEMBLY,
		'REQUIRED_OPTION_MISSING':               REQUIRED_OPTION_MISSING,
		'IP_HEADER_BAD':                         IP_HEADER_BAD,
	}.items())

	# names would have non-unique keys

	def __str__ (self):
		return '%d' % int(self)



"""
link.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import unpack

from exabgp.bgp.message.update.nlri.bgpls.nlri import BGPLS
from exabgp.bgp.message.update.nlri.bgpls.nlri import PROTO_CODES
from exabgp.bgp.message.update.nlri.bgpls.tlvs.linkid import LinkIdentifier
from exabgp.bgp.message.update.nlri.bgpls.tlvs.ifaceaddr import IfaceAddr
from exabgp.bgp.message.update.nlri.bgpls.tlvs.neighaddr import NeighAddr
from exabgp.bgp.message.update.nlri.bgpls.tlvs.node import NodeDescriptor


#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+
#     |  Protocol-ID  |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                           Identifier                          |
#     |                            (64 bits)                          |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //               Local Node Descriptors (variable)             //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //               Remote Node Descriptors (variable)            //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //                  Link Descriptors (variable)                //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#                     Figure 8: The Link NLRI format
#
#   +-----------+---------------------+--------------+------------------+
#   |  TLV Code | Description         |  IS-IS TLV   | Reference        |
#   |   Point   |                     |   /Sub-TLV   | (RFC/Section)    |
#   +-----------+---------------------+--------------+------------------+
#   |    258    | Link Local/Remote   |     22/4     | [RFC5307]/1.1    |
#   |           | Identifiers         |              |                  |
#   |    259    | IPv4 interface      |     22/6     | [RFC5305]/3.2    |
#   |           | address             |              |                  |
#   |    260    | IPv4 neighbor       |     22/8     | [RFC5305]/3.3    |
#   |           | address             |              |                  |
#   |    261    | IPv6 interface      |    22/12     | [RFC6119]/4.2    |
#   |           | address             |              |                  |
#   |    262    | IPv6 neighbor       |    22/13     | [RFC6119]/4.3    |
#   |           | address             |              |                  |
#   |    263    | Multi-Topology      |     ---      | Section 3.2.1.5  |
#   |           | Identifier          |              |                  |
#   +-----------+---------------------+--------------+------------------+


@BGPLS.register
class LINK(BGPLS):
	CODE = 2
	NAME = "Link NLRI"
	SHORT_NAME = "Link"

	def __init__ (
			self,domain,proto_id,local_node,
			remote_node,neigh_addrs=None,iface_addrs=None,
			packed=None,link_ids=None,nexthop=None,
			action=None, route_d=None, addpath=None):
		BGPLS.__init__(self,action,addpath)
		self.domain = domain
		self.proto_id = proto_id
		self.local_node = local_node
		self.remote_node = remote_node
		self.neigh_addrs = neigh_addrs
		self.iface_addrs = iface_addrs
		self.link_ids = link_ids
		self.nexthop = nexthop
		self.route_d = route_d
		self._pack = packed

	@classmethod
	def unpack_nlri (cls, data, rd):
		proto_id = unpack('!B',data[0:1])[0]
		iface_addrs = []
		neigh_addrs = []
		link_identifiers = []
		remote_node = []
		local_node = []
		if proto_id not in PROTO_CODES.keys():
			raise Exception('Protocol-ID {} is not valid'.format(proto_id))
		domain = unpack('!Q',data[1:9])[0]
		tlvs = data[9:]

		while tlvs:
			tlv_type, tlv_length = unpack('!HH', tlvs[:4])
			if tlv_type == 256:
				values = tlvs[4: 4 + tlv_length]
				local_node = []
				while values:
					# Unpack Local Node Descriptor Sub-TLVs
					# We pass proto_id as TLV interpretation
					# follows IGP type
					node, left = NodeDescriptor.unpack(values, proto_id)
					local_node.append(node)
					if left == tlvs:
						raise RuntimeError("sub-calls should consume data")
					values = left
				tlvs = tlvs[4 + tlv_length:]
				continue
			elif tlv_type == 257:
				# Remote Node Descriptor
				values = tlvs[4: 4 + tlv_length]
				remote_node = []
				while values:
					node, left = NodeDescriptor.unpack(values, proto_id)
					remote_node.append(node)
					if left == tlvs:
						raise RuntimeError("sub-calls should consume data")
					values = left
				tlvs = tlvs[4 + tlv_length:]
				continue
			elif tlv_type == 258:
				# Link Local/Remote identifiers
				value = tlvs[4: 4 + 8]
				link_identifiers = LinkIdentifier.unpack(value)
				tlvs = tlvs[4 + 8:]
				continue
			elif tlv_type in [259,261]:
				# IPv{4,6} Interface Address
				value = tlvs[4: 4 + tlv_length]
				iface_addrs.append(IfaceAddr.unpack(value))
				tlvs = tlvs[4 + tlv_length:]
				continue
			elif tlv_type in [260,262]:
				# IPv{4,6} Neighbor Address
				value = tlvs[4: 4 + tlv_length]
				neigh_addrs.append(NeighAddr.unpack(value))
				tlvs = tlvs[4 + tlv_length:]
				continue

		return cls(
			domain=domain,proto_id=proto_id,
			local_node=local_node,remote_node=remote_node,
			neigh_addrs=neigh_addrs,iface_addrs=iface_addrs,
			link_ids=link_identifiers,route_d=rd,packed=data
		)

	def __eq__ (self, other):
		return \
			isinstance(other, LINK) and \
			self.CODE == other.CODE and \
			self.domain == other.domain and \
			self.proto_id == other.proto_id and \
			self.route_d == other.route_d

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __str__ (self):
		return self.json()

	def __hash__ (self):
		return hash((self))

	def pack (self,negotiated=None):
		return self._pack

	def json (self, compact=None):
		local = ', '.join(d.json() for d in self.local_node)
		remote = ', '.join(d.json() for d in self.remote_node)
		interface_addrs = ', '.join(d.json() for d in self.iface_addrs)
		neighbor_addrs = ', '.join(d.json() for d in self.neigh_addrs)
		content = '"ls-nlri-type": 2, '
		content += '"l3-routing-topology": %d, ' % int(self.domain)
		content += '"protocol-id": %d, ' % int(self.proto_id)
		content += '"local-node-descriptors": { %s }, ' % local
		content += '"remote-node-descriptors": { %s }, ' % remote
		content += '"interface-address": { %s }, ' % interface_addrs
		content += '"neighbor-address": { %s }' % neighbor_addrs
		if self.link_ids:
			links = ', '.join(d.json() for d in self.link_ids)
			content += '" ,link-identifiers": { %s }' % links
		if self.route_d:
			content += ", { %s }" % self.route_d.json()
		return '{ %s }' % content



# encoding: utf-8
"""
command/neighbor.py

Created by Thomas Mangin on 2017-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from datetime import timedelta

from exabgp.reactor.api.command.command import Command
from exabgp.reactor.api.command.limit import match_neighbor
from exabgp.reactor.api.command.limit import extract_neighbors

from exabgp.reactor.api.response.answer import Answer


def register_neighbor ():
	pass


def _en (value):
	if value is None:
		return 'n/a'
	return 'enabled' if value else 'disabled'


def _pr (value):
	if value is None:
		return 'n/a'
	return '%s' % value


class Neighbor (object):
	extensive_kv = '   %-20s %15s %15s %15s'
	extensive_template = """\
Neighbor %(peer-address)s

	Session                         Local
%(local-address)s
%(state)s
%(duration)s

	Setup                           Local          Remote
%(as)s
%(id)s
%(hold)s

	Capability                      Local          Remote
%(capabilities)s

	Families                        Local          Remote        Add-Path
%(families)s

	Message Statistic                Sent        Received
%(messages)s
""".replace('\t','  ')

	summary_header   = 'Peer            AS        up/down state       |     #sent     #recvd'
	summary_template = '%-15s %-7s %9s %-12s %10d %10d'

	@classmethod
	def extensive (cls,answer):
		if answer['duration']:
			duration = cls.extensive_kv % ('up for', timedelta(seconds=answer['duration']), '', '')
		else:
			duration = cls.extensive_kv % ('down for', timedelta(seconds=answer['down']), '', '')
		formated = {
			'peer-address':  answer['peer-address'],
			'local-address': cls.extensive_kv % ('local',answer['local-address'],'',''),
			'state':         cls.extensive_kv % ('state',answer['state'],'',''),
			'duration':      duration,
			'as':            cls.extensive_kv % ('AS',answer['local-as'],_pr(answer['peer-as']),''),
			'id':            cls.extensive_kv % ('ID',answer['local-id'],_pr(answer['peer-id']),''),
			'hold':          cls.extensive_kv % ('hold-time',answer['local-hold'],_pr(answer['peer-hold']),''),
			'capabilities':  '\n'.join(cls.extensive_kv % ('%s:' % k, _en(l), _en(p), '') for k,(l,p) in answer['capabilities'].items()),
			'families':      '\n'.join(cls.extensive_kv % ('%s %s:' % (a,s), _en(l), _en(r), _en(p)) for (a,s),(l,r,p) in answer['families'].items()),
			'messages':      '\n'.join(cls.extensive_kv % ('%s:' % k, str(s), str(r), '') for k,(s,r) in answer['messages'].items()),
		}

		return cls.extensive_template % formated

	@classmethod
	def summary (cls,answer):
		return cls.summary_template % (
			answer['peer-address'],
			_pr(answer['peer-as']),
			timedelta(seconds=answer['duration']) if answer['duration'] else 'down',
			answer['state'].lower(),
			answer['messages']['update'][0],
			answer['messages']['update'][1]
		)


@Command.register('text','teardown',True)
def teardown (self, reactor, service, line):
	try:
		descriptions,line = extract_neighbors(line)
		_,code = line.split(' ',1)
		for key in reactor.peers:
			for description in descriptions:
				if match_neighbor(description,key):
					reactor.peers[key].teardown(int(code))
					self.log_message('teardown scheduled for %s' % ' '.join(description))
		reactor.processes.answer_done(service)
		return True
	except ValueError:
		reactor.processes.answer(service,Answer.error)
		return False
	except IndexError:
		reactor.processes.answer(service,Answer.error)
		return False


@Command.register('text', 'show neighbor', False, ['summary', 'extensive', 'configuration'])
def show_neighbor (self, reactor, service, command):
	words = command.split()

	extensive = 'extensive' in words
	configuration = 'configuration' in words
	summary = 'summary' in words

	if summary:
		words.remove('summary')
	if extensive:
		words.remove('extensive')
	if configuration:
		words.remove('configuration')

	limit = words[-1] if words[-1] != 'neighbor' else ''

	def callback_configuration ():
		for neighbor_name in reactor.configuration.neighbors.keys():
			neighbor = reactor.configuration.neighbors.get(neighbor_name,None)
			if not neighbor:
				continue
			if limit and limit not in neighbor_name:
				continue
			for line in str(neighbor).split('\n'):
				reactor.processes.answer(service,line)
				yield True
		reactor.processes.answer_done(service)

	def callback_extensive ():
		for peer_name in reactor.peers.keys():
			peer = reactor.peers.get(peer_name,None)
			if not peer:
				continue
			if limit and limit not in peer.neighbor.name():
				continue
			for line in Neighbor.extensive(peer.cli_data()).split('\n'):
				reactor.processes.answer(service,line)
				yield True
		reactor.processes.answer_done(service)

	def callback_summary ():
		reactor.processes.answer(service,Neighbor.summary_header)
		for peer_name in reactor.peers.keys():
			peer = reactor.peers.get(peer_name,None)
			if not peer:
				continue
			if limit and limit != str(peer.neighbor.peer_address):
				continue
			for line in Neighbor.summary(peer.cli_data()).split('\n'):
				reactor.processes.answer(service,line)
				yield True
		reactor.processes.answer_done(service)

	if summary:
		reactor.asynchronous.schedule(service,command,callback_summary())
		return True

	if extensive:
		reactor.asynchronous.schedule(service, command, callback_extensive())
		return True

	if configuration:
		reactor.asynchronous.schedule(service, command, callback_configuration())
		return True

	reactor.processes.answer(service,'please specify summary, extensive or configuration')
	reactor.processes.answer(service,'you can filter by peer ip address adding it after the word neighbor')
	reactor.processes.answer_done(service)



# encoding: utf-8
"""
neighbor.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os
import uuid

from collections import deque

# collections.counter is python2.7 only ..
try:
	from collections import Counter
except ImportError:
	from exabgp.vendoring.counter import Counter

from exabgp.protocol.family import AFI

from exabgp.bgp.message import Message
from exabgp.bgp.message.open.capability import AddPath

from exabgp.rib import RIB


# The definition of a neighbor (from reading the configuration)
class Neighbor (object):
	_GLOBAL = {'uid': 1}

	def __init__ (self):
		# self.logger should not be used here as long as we do use deepcopy as it contains a Lock
		self.description = None
		self.router_id = None
		self.host_name = None
		self.domain_name = None
		self.local_address = None
		self.range_size = 1
		# local_address uses auto discovery
		self.auto_discovery = False
		self.peer_address = None
		self.peer_as = None
		self.local_as = None
		self.hold_time = None
		self.asn4 = None
		self.add_path = None
		self.md5_password = None
		self.md5_base64 = False
		self.md5_ip = None
		self.ttl_in = None
		self.ttl_out = None
		self.group_updates = None
		self.flush = None
		self.adj_rib_in = None
		self.adj_rib_out = None

		self.manual_eor = False

		self.api = None  # XXX: not scriptable - is replaced outside the class
		# passive indicate that we do not establish outgoing connections
		self.passive = False
		# the port to listen on ( zero mean that we do not listen )
		self.listen = 0
		# the port to connect to
		self.connect = 0

		# was this Neighbor generated from a range
		self.generated = False

		# capability
		self.route_refresh = False
		self.graceful_restart = False
		self.multisession = None
		self.add_path = None
		self.aigp = None

		self._families = []
		self._addpaths = []
		self.rib = None

		# The routes we have parsed from the configuration
		self.changes = []
		# On signal update, the previous routes so we can compare what changed
		self.backup_changes = []

		self.operational = None
		self.eor = deque()
		self.asm = dict()

		self.messages = deque()
		self.refresh = deque()

		self.counter = Counter()
		# It is possible to :
		# - have multiple exabgp toward one peer on the same host ( use of pid )
		# - have more than once connection toward a peer
		# - each connection has it own neihgbor (hence why identificator is not in Protocol)
		self.uid = '%d' % self._GLOBAL['uid']
		self._GLOBAL['uid'] += 1

	def id (self):
		return 'neighbor-%s' % self.uid

	def make_rib (self):
		self.rib = RIB(self.name(),self.adj_rib_in,self.adj_rib_out,self._families)

	# will resend all the routes once we reconnect
	def reset_rib (self):
		self.rib.reset()
		self.messages = deque()
		self.refresh = deque()

	# back to square one, all the routes are removed
	def clear_rib (self):
		self.rib.clear()
		self.messages = deque()
		self.refresh = deque()

	def name (self):
		if self.multisession:
			session = '/'.join("%s-%s" % (afi.name(),safi.name()) for (afi,safi) in self.families())
		else:
			session = 'in-open'
		return "neighbor %s local-ip %s local-as %s peer-as %s router-id %s family-allowed %s" % (
			self.peer_address,
			self.local_address if self.peer_address is not None else 'auto',
			self.local_as if self.local_as is not None else 'auto',
			self.peer_as if self.peer_as is not None else 'auto',
			self.router_id,
			session
		)

	def families (self):
		# this list() is important .. as we use the function to modify self._families
		return list(self._families)

	def addpaths (self):
		# this list() is important .. as we use the function to modify self._families
		return list(self._addpaths)

	def add_family (self, family):
		# the families MUST be sorted for neighbor indexing name to be predictable for API users
		# this list() is important .. as we use the function to modify self._families
		if family not in self.families():
			afi,safi = family
			d = dict()
			d[afi] = [safi,]
			for afi,safi in self._families:
				d.setdefault(afi,[]).append(safi)
			self._families = [(afi,safi) for afi in sorted(d) for safi in sorted(d[afi])]

	def add_addpath (self, family):
		# the families MUST be sorted for neighbor indexing name to be predictable for API users
		# this list() is important .. as we use the function to modify self._families
		if family not in self.addpaths():
			afi,safi = family
			d = dict()
			d[afi] = [safi,]
			for afi,safi in self._addpaths:
				d.setdefault(afi,[]).append(safi)
			self._addpaths = [(afi,safi) for afi in sorted(d) for safi in sorted(d[afi])]

	def remove_family (self, family):
		if family in self.families():
			self._families.remove(family)

	def remove_addpath (self, family):
		if family in self.addpaths():
			self._addpaths.remove(family)

	def missing (self):
		if self.local_address is None and not self.auto_discovery:
			return 'local-address'
		if self.listen > 0 and self.auto_discovery:
			return 'local-address'
		if self.peer_address is None:
			return 'peer-address'
		if self.auto_discovery and not self.router_id:
			return 'router-id'
		if self.peer_address.afi == AFI.ipv6 and not self.router_id:
			return 'router-id'
		return ''

	# This function only compares the neighbor BUT NOT ITS ROUTES
	def __eq__ (self, other):
		# Comparing local_address is skipped in the case where either
		# peer is configured to auto discover its local address. In
		# this case it can happen that one local_address is None and
		# the other one will be set to the auto disocvered IP address.
		auto_discovery = self.auto_discovery or other.auto_discovery
		return \
			self.router_id == other.router_id and \
			(auto_discovery or self.local_address == other.local_address) and \
			self.auto_discovery == other.auto_discovery and \
			self.local_as == other.local_as and \
			self.peer_address == other.peer_address and \
			self.peer_as == other.peer_as and \
			self.passive == other.passive and \
			self.listen == other.listen and \
			self.connect == other.connect and \
			self.hold_time == other.hold_time and \
			self.host_name == other.host_name and \
			self.domain_name == other.domain_name and \
			self.md5_password == other.md5_password and \
			self.md5_ip == other.md5_ip and \
			self.ttl_in == other.ttl_in and \
			self.ttl_out == other.ttl_out and \
			self.route_refresh == other.route_refresh and \
			self.graceful_restart == other.graceful_restart and \
			self.multisession == other.multisession and \
			self.add_path == other.add_path and \
			self.operational == other.operational and \
			self.group_updates == other.group_updates and \
			self.flush == other.flush and \
			self.adj_rib_in == other.adj_rib_in and \
			self.adj_rib_out == other.adj_rib_out and \
			self.families() == other.families()

	def __ne__ (self, other):
		return not self.__eq__(other)

	def string (self, with_changes=True):
		changes = ''
		if with_changes:
			changes += '\nstatic { '
			for changes in self.rib.incoming.queued_changes():
				changes += '\n\t\t%s' % changes.extensive()
			changes += '\n}'

		families = ''
		for afi,safi in self.families():
			families += '\n\t\t%s %s;' % (afi.name(),safi.name())

		addpaths = ''
		for afi,safi in self.addpaths():
			addpaths += '\n\t\t%s %s;' % (afi.name(),safi.name())

		codes = Message.CODE

		_extension_global = {
			'neighbor-changes': 'neighbor-changes',
			'negotiated':       'negotiated',
			'fsm':              'fsm',
			'signal':           'signal',
		}

		_extension_receive = {
			'receive-packets':                        'packets',
			'receive-parsed':                         'parsed',
			'receive-consolidate':                    'consolidate',
			'receive-%s' % codes.NOTIFICATION.SHORT:  'notification',
			'receive-%s' % codes.OPEN.SHORT:          'open',
			'receive-%s' % codes.KEEPALIVE.SHORT:     'keepalive',
			'receive-%s' % codes.UPDATE.SHORT:        'update',
			'receive-%s' % codes.ROUTE_REFRESH.SHORT: 'refresh',
			'receive-%s' % codes.OPERATIONAL.SHORT:   'operational',
		}

		_extension_send = {
			'send-packets':                        'packets',
			'send-parsed':                         'parsed',
			'send-consolidate':                    'consolidate',
			'send-%s' % codes.NOTIFICATION.SHORT:  'notification',
			'send-%s' % codes.OPEN.SHORT:          'open',
			'send-%s' % codes.KEEPALIVE.SHORT:     'keepalive',
			'send-%s' % codes.UPDATE.SHORT:        'update',
			'send-%s' % codes.ROUTE_REFRESH.SHORT: 'refresh',
			'send-%s' % codes.OPERATIONAL.SHORT:   'operational',
		}

		apis = ''

		for process in self.api.get('processes',[]):
			_global = []
			_receive = []
			_send = []

			for api,name in _extension_global.items():
				_global.extend(['\t\t%s;\n' % name,] if process in self.api[api] else [])

			for api,name in _extension_receive.items():
				_receive.extend(['\t\t\t%s;\n' % name,] if process in self.api[api] else [])

			for api,name in _extension_send.items():
				_send.extend(['\t\t\t%s;\n' % name,] if process in self.api[api] else [])

			_api  = '\tapi {\n'
			_api += '\t\tprocesses [ %s ];\n' % process
			_api += ''.join(_global)
			if _receive:
				_api += '\t\treceive {\n'
				_api += ''.join(_receive)
				_api += '\t\t}\n'
			if _send:
				_api += '\t\tsend {\n'
				_api += ''.join(_send)
				_api += '\t\t}\n'
			_api += '\t}\n'

			apis += _api

		returned = \
			'neighbor %s {\n' \
			'\tdescription "%s";\n' \
			'\trouter-id %s;\n' \
			'\thost-name %s;\n' \
			'\tdomain-name %s;\n' \
			'\tlocal-address %s;\n' \
			'\tlocal-as %s;\n' \
			'\tpeer-as %s;\n' \
			'\thold-time %s;\n' \
			'\tmanual-eor %s;\n' \
			'%s%s%s%s%s%s%s%s%s%s%s\n' \
			'\tcapability {\n' \
			'%s%s%s%s%s%s%s%s\t}\n' \
			'\tfamily {%s\n' \
			'\t}\n' \
			'\tadd-path {%s\n' \
			'\t}\n' \
			'%s' \
			'%s' \
			'}' % (
				self.peer_address,
				self.description,
				self.router_id,
				self.host_name,
				self.domain_name,
				self.local_address if not self.auto_discovery else 'auto',
				self.local_as,
				self.peer_as,
				self.hold_time,
				'true' if self.manual_eor else 'false',
				'\n\tpassive %s;\n' % ('true' if self.passive else 'false'),
				'\n\tlisten %d;\n' % self.listen if self.listen else '',
				'\n\tconnect %d;\n' % self.connect if self.connect else '',
				'\tgroup-updates %s;\n' % ('true' if self.group_updates else 'false'),
				'\tauto-flush %s;\n' % ('true' if self.flush else 'false'),
				'\tadj-rib-in %s;\n' % ('true' if self.adj_rib_in else 'false'),
				'\tadj-rib-out %s;\n' % ('true' if self.adj_rib_out else 'false'),
				'\tmd5-password "%s";\n' % self.md5_password if self.md5_password else '',
				'\tmd5-base64 %s;\n' % ('true' if self.md5_base64 is True else 'false' if self.md5_base64 is False else 'auto'),
				'\tmd5-ip "%s";\n' % self.md5_ip if not self.auto_discovery else '',
				'\toutgoing-ttl %s;\n' % self.ttl_out if self.ttl_out else '',
				'\tincoming-ttl %s;\n' % self.ttl_in if self.ttl_in else '',
				'\t\tasn4 %s;\n' % ('enable' if self.asn4 else 'disable'),
				'\t\troute-refresh %s;\n' % ('enable' if self.route_refresh else 'disable'),
				'\t\tgraceful-restart %s;\n' % (self.graceful_restart if self.graceful_restart else 'disable'),
				'\t\tadd-path %s;\n' % (AddPath.string[self.add_path] if self.add_path else 'disable'),
				'\t\tmulti-session %s;\n' % ('enable' if self.multisession else 'disable'),
				'\t\toperational %s;\n' % ('enable' if self.operational else 'disable'),
				'\t\taigp %s;\n' % ('enable' if self.aigp else 'disable'),
				families,
				addpaths,
				apis,
				changes
			)

		# '\t\treceive {\n%s\t\t}\n' % receive if receive else '',
		# '\t\tsend {\n%s\t\t}\n' % send if send else '',
		return returned.replace('\t','  ')

	def __str__ (self):
		return self.string(False)



# encoding: utf-8
"""
decoder/__init__.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.core.format import formated
from exabgp.configuration.operational.parser import operational

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.protocol.family import Family

from exabgp.bgp.message.refresh import RouteRefresh

from exabgp.logger import Logger
from exabgp.reactor.api.command import Command
from exabgp.configuration.configuration import Configuration

from exabgp.reactor.api.response.answer import Answer

# ======================================================================= Parser
#


class API (Command):
	def __init__ (self, reactor):
		self.reactor = reactor
		self.logger = Logger()
		self.configuration = Configuration([])

	def log_message (self, message, level='INFO'):
		self.logger.notice(message,'api',level)

	def log_failure (self, message, level='ERR'):
		error = str(self.configuration.tokeniser.error)
		report = '%s\nreason: %s' % (message, error) if error else message
		self.logger.error(report,'api',level)

	def text (self, reactor, service, command):
		for registered in self.functions:
			if registered == command or registered + ' ' in command:
				return self.callback['text'][registered](self,reactor,service,command)
		reactor.processes.answer(service,Answer.error)
		self.logger.warning('command from process not understood : %s' % command,'api')
		return False

	def api_route (self, command):
		action, line = command.split(' ',1)

		self.configuration.static.clear()
		if not self.configuration.partial('static',line):
			return []

		if self.configuration.scope.location():
			return []

		self.configuration.scope.to_context()
		changes = self.configuration.scope.pop_routes()
		return changes

	def api_flow (self, command):
		action, flow, line = command.split(' ',2)

		self.configuration.flow.clear()
		if not self.configuration.partial('flow',line):
			return []

		if self.configuration.scope.location():
			return []

		self.configuration.scope.to_context()
		changes = self.configuration.scope.pop_routes()
		return changes

	def api_vpls (self, command):
		action, line = command.split(' ',1)

		self.configuration.l2vpn.clear()
		if not self.configuration.partial('l2vpn',line):
			return []

		self.configuration.scope.to_context()
		changes = self.configuration.scope.pop('l2vpn')
		return changes

	def api_attributes (self, command, peers):
		action, line = command.split(' ',1)

		self.configuration.static.clear()
		if not self.configuration.partial('static',line):
			return []

		self.configuration.scope.to_context()
		changes = self.configuration.scope.pop_routes()
		return changes

	def api_refresh (self, command):
		tokens = formated(command).split(' ')[2:]
		if len(tokens) != 2:
			return False
		afi = AFI.value(tokens.pop(0))
		safi = SAFI.value(tokens.pop(0))
		if afi is None or safi is None:
			return False
		return [RouteRefresh(afi,safi)]

	def api_eor (self, command):
		tokens = formated(command).split(' ')[2:]
		number = len(tokens)

		if not number:
			return Family(1,1)

		if number != 2:
			return False

		afi = AFI.fromString(tokens[0])
		if afi == AFI.undefined:
			return False

		safi = SAFI.fromString(tokens[1])
		if safi == SAFI.undefined:
			return False

		return Family(afi,safi)

	def api_operational (self, command):
		tokens = formated(command).split(' ')

		op = tokens[1].lower()
		what = tokens[2].lower()

		if op != 'operational':
			return False

		self.configuration.tokeniser.iterate.replenish(tokens[3:])
		# None or a class
		return operational(what,self.configuration.tokeniser.iterate)



# encoding: utf-8
"""
line/watchdog.py

Created by Thomas Mangin on 2017-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.reactor.api.command.command import Command


def register_watchdog ():
	pass


@Command.register('text','announce watchdog')
def announce_watchdog (self, reactor, service, line):
	def callback (name):
		# XXX: move into Action
		for neighbor_name in reactor.configuration.neighbors.keys():
			neighbor = reactor.configuration.neighbors.get(neighbor_name, None)
			if not neighbor:
				continue
			neighbor.rib.outgoing.announce_watchdog(name)
			yield False

		reactor.processes.answer_done(service)

	try:
		name = line.split(' ')[2]
	except IndexError:
		name = service
	reactor.asynchronous.schedule(service, line, callback(name))
	return True


@Command.register('text','withdraw watchdog')
def withdraw_watchdog (self, reactor, service, line):
	def callback (name):
		# XXX: move into Action
		for neighbor_name in reactor.configuration.neighbors.keys():
			neighbor = reactor.configuration.neighbors.get(neighbor_name, None)
			if not neighbor:
				continue
			neighbor.rib.outgoing.withdraw_watchdog(name)
			yield False

		reactor.processes.answer_done(service)

	try:
		name = line.split(' ')[2]
	except IndexError:
		name = service
	reactor.asynchronous.schedule(service, line, callback(name))
	return True



# encoding: utf-8
"""
scope.py

Created by Stephane Litkowski on 2017-02-24.
"""

from exabgp.configuration.core import Section

from exabgp.configuration.flow.parser import interface_set


class ParseFlowScope (Section):
	definition = [
		'interface-set transitive:input:1234:1234'
	]

	syntax = \
		'scope {\n' \
		'  %s;\n' \
		'}' % ';\n  '.join(definition)

	known = {
		'interface-set':	interface_set,
	}

	# 'community','extended-community'

	action = {
		'interface-set':       'attribute-add',
	}

	name = 'flow/scope'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		pass

	def pre (self):
		return True

	def post (self):
		return True



# encoding: utf-8
"""
refresh.py

Created by Thomas Mangin on 2012-07-19.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import unpack
from struct import error

from exabgp.util import character
from exabgp.util import concat_bytes
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message.message import Message
from exabgp.bgp.message.notification import Notify

# =================================================================== Notification
# A Notification received from our peer.
# RFC 4271 Section 4.5


class Reserved (int):
	def __str__ (self):
		if self == 0:
			return 'query'
		if self == 1:
			return 'begin'
		if self == 2:
			return 'end'
		return 'invalid'


@Message.register
class RouteRefresh (Message):
	ID = Message.CODE.ROUTE_REFRESH
	TYPE = character(Message.CODE.ROUTE_REFRESH)

	request = 0
	start = 1
	end = 2

	def __init__ (self, afi, safi, reserved=0):
		self.afi = AFI.create(afi)
		self.safi = SAFI.create(safi)
		self.reserved = Reserved(reserved)

	def message (self,negotiated=None):
		return self._message(concat_bytes(self.afi.pack(),character(self.reserved),self.safi.pack()))

	def __str__ (self):
		return "REFRESH"

	def extensive (self):
		return 'route refresh %s/%d/%s' % (self.afi,self.reserved,self.safi)

	# XXX: Check how we get this data into the RR
	# def families (self):
	# 	return self._families[:]

	@classmethod
	def unpack_message (cls, data, _):
		try:
			afi,reserved,safi = unpack('!HBB',data)
		except error:
			raise Notify(7,1,'invalid route-refresh message')
		if reserved not in (0,1,2):
			raise Notify(7,2,'invalid route-refresh message subtype')
		return RouteRefresh(afi,safi,reserved)

	def __eq__ (self, other):
		if not isinstance(other, RouteRefresh):
			return False
		if self.afi != other.afi:
			return False
		if self.safi != other.safi:
			return False
		if self.reserved != other.reserved:
			return False
		return True

	def __ne__ (self, other):
		return not self.__eq__(other)



# encoding: utf-8
"""
reactor/interrupt.py

Created by Thomas Mangin on 2017-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import signal

from exabgp.logger import Logger


class Signal (object):
	NONE        = 0
	SHUTDOWN    = 1
	RESTART     = 2
	RELOAD      = 4
	FULL_RELOAD = 8

	def __init__ (self):
		self.logger = Logger()
		self.received = self.NONE
		self.number = 0
		self.rearm()

	def rearm (self):
		self.received = Signal.NONE
		self.number = 0

		signal.signal(signal.SIGTERM, self.sigterm)
		signal.signal(signal.SIGHUP, self.sighup)
		signal.signal(signal.SIGALRM, self.sigalrm)
		signal.signal(signal.SIGUSR1, self.sigusr1)
		signal.signal(signal.SIGUSR2, self.sigusr2)

	def sigterm (self, signum, frame):
		self.logger.critical('SIGTERM received','reactor')
		if self.received:
			self.logger.critical('ignoring - still handling previous signal','reactor')
			return
		self.logger.critical('scheduling shutdown','reactor')
		self.received = self.SHUTDOWN
		self.number = signum

	def sighup (self, signum, frame):
		self.logger.critical('SIGHUP received','reactor')
		if self.received:
			self.logger.critical('ignoring - still handling previous signal','reactor')
			return
		self.logger.critical('scheduling shutdown','reactor')
		self.received = self.SHUTDOWN
		self.number = signum

	def sigalrm (self, signum, frame):
		self.logger.critical('SIGALRM received','reactor')
		if self.received:
			self.logger.critical('ignoring - still handling previous signal','reactor')
			return
		self.logger.critical('scheduling restart','reactor')
		self.received = self.RESTART
		self.number = signum

	def sigusr1 (self, signum, frame):
		self.logger.critical('SIGUSR1 received','reactor')
		if self.received:
			self.logger.critical('ignoring - still handling previous signal','reactor')
			return
		self.logger.critical('scheduling reload of configuration','reactor')
		self.received = self.RELOAD
		self.number = signum

	def sigusr2 (self, signum, frame):
		self.logger.critical('SIGUSR2 received','reactor')
		if self.received:
			self.logger.critical('ignoring - still handling previous signal','reactor')
			return
		self.logger.critical('scheduling reload of configuration and processes','reactor')
		self.received = self.FULL_RELOAD
		self.number = signum



"""
control.py

Created by Thomas Mangin on 2015-01-13.
Copyright (c) 2015-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os
import sys
import fcntl
import stat
import time
import signal
import select
import socket
import traceback
from collections import deque

from exabgp.util import str_ascii
from exabgp.reactor.network.error import error

kb = 1024
mb = kb*1024


def env (app, section, name, default):
	r = os.environ.get('%s.%s.%s' % (app,section,name), None)
	if r is None:
		r = os.environ.get('%s_%s_%s' % (app,section,name), None)
	if r is None:
		return default
	return r


def check_fifo (name):
	try:
		if not stat.S_ISFIFO(os.stat(name).st_mode):
			sys.stdout.write('error: a file exist which is not a named pipe (%s)\n' % os.path.abspath(name))
			return False

		if not os.access(name,os.R_OK):
			sys.stdout.write('error: a named pipe exists and we can not read/write to it (%s)\n' % os.path.abspath(name))
			return False
		return True
	except OSError:
		sys.stdout.write('error: could not create the named pipe %s\n' % os.path.abspath(name))
		return False
	except IOError:
		sys.stdout.write('error: could not access/delete the named pipe %s\n' % os.path.abspath(name))
		sys.stdout.flush()
	except socket.error:
		sys.stdout.write('error: could not write on the named pipe %s\n' % os.path.abspath(name))
		sys.stdout.flush()


class Control (object):
	terminating = False

	def __init__ (self, location):
		self.send = location + env('exabgp','api','pipename','exabgp') + '.out'
		self.recv = location + env('exabgp','api','pipename','exabgp') + '.in'
		self.r_pipe = None

	def init (self):
		# obviously this is vulnerable to race conditions ... if an attacker can create fifo in the folder

		if not check_fifo(self.recv):
			self.terminate()
			sys.exit(1)

		if not check_fifo(self.send):
			self.terminate()
			sys.exit(1)

		signal.signal(signal.SIGINT, self.terminate)
		signal.signal(signal.SIGTERM, self.terminate)
		return True

	def cleanup (self):
		def _close (pipe):
			if self.r_pipe:
				try:
					os.close(pipe)
				except (OSError,IOError,TypeError):
					pass

		_close(self.r_pipe)

	def terminate (self,ignore=None,me=None):
		# if the named pipe is open, and remove_fifo called
		# do not ignore a second signal
		if self.terminating:
			sys.exit(1)
		self.terminating = True

		self.cleanup()

		# def _remove_fifo (name):
		# 	# If we got two signal, time to stop trying nicely as we can not delete a file which is open
		# 	try:
		# 		if os.path.exists(name):
		# 			os.remove(name)
		# 	except IOError:
		# 		sys.stdout.write('error: could not remove current named pipe (%s)\n' % os.path.abspath(name))
		# 		sys.stdout.flush()

		# self._remove_fifo(self.recv)
		# self._remove_fifo(self.send)

	def read_on (self,reading):
		sleep_time = 1.0
		fault_time = 0.9 * sleep_time
		now = time.time()

		try:
			ready,_,_ = select.select(reading,[],[],sleep_time)
		except select.error as e:
			if e.args[0] in error.block:
				return []
			sys.exit(1)  # Unknow error, ending

		# select stoped before the 1 second but we have no data
		# the PIPE with ExaBGP is dead
		elapsed = time.time() - now
		if not reading and elapsed < fault_time:
			sys.exit(1)
		return ready

	def no_buffer (self,fd):
		mfl = fcntl.fcntl(fd, fcntl.F_GETFL)
		mfl |= os.O_SYNC
		fcntl.fcntl(fd, fcntl.F_SETFL, mfl)

	def loop (self):
		try:
			self.r_pipe = os.open(self.recv, os.O_RDWR | os.O_NONBLOCK | os.O_EXCL)
		except OSError:
			self.terminate()

		standard_in = sys.stdin.fileno()
		standard_out = sys.stdout.fileno()

		def monitor (function):
			def wrapper (*args):
				# print >> sys.stderr, "%s(%s)" % (function.func_name,','.join([str(_).replace('\n','\\n') for _ in args]))
				r = function(*args)
				# print >> sys.stderr, "%s -> %s" % (function.func_name,str(r))
				return r
			return wrapper

		@monitor
		def std_reader (number):
			try:
				return os.read(standard_in,number)
			except OSError as exc:
				if exc.errno in error.block:
					return ''
				sys.exit(1)

		@monitor
		def std_writer (line):
			try:
				return os.write(standard_out,line)
			except OSError as exc:
				if exc.errno in error.block:
					return 0
				sys.exit(1)

		@monitor
		def fifo_reader (number):
			try:
				return os.read(self.r_pipe,number)
			except OSError as exc:
				if exc.errno in error.block:
					return ''
				sys.exit(1)

		@monitor
		def fifo_writer (line):
			pipe,nb = None,0
			try:
				pipe = os.open(self.send, os.O_WRONLY | os.O_NONBLOCK | os.O_EXCL)
				self.no_buffer(pipe)
			except OSError:
				time.sleep(0.05)
				return 0
			if pipe is not None:
				try:
					nb = os.write(pipe,line)
				except OSError:
					pass
				try:
					os.close(pipe)
				except OSError:
					pass
			return nb

		read = {
			standard_in: std_reader,
			self.r_pipe: fifo_reader,
		}

		write = {
			standard_in: fifo_writer,
			self.r_pipe: std_writer,
		}

		backlog = {
			standard_in: deque(),
			self.r_pipe: deque(),
		}

		store = {
			standard_in: b'',
			self.r_pipe: b'',
		}

		def consume (source):
			if not backlog[source] and b'\n' not in store[source]:
				store[source] += read[source](1024)
			else:
				backlog[source].append(read[source](1024))
				# assuming a route takes 80 chars, 100 Mb is over 1Millions routes
				# something is really wrong if it was not consummed
				if len(backlog) > 100*mb:
					sys.stderr.write('using too much memory - exiting')
					sys.exit(1)

		reading = [standard_in, self.r_pipe]

		while True:
			ready = self.read_on(reading)

			# command from user
			if self.r_pipe in ready:
				consume(self.r_pipe)
			if standard_in in ready:
				consume(standard_in)

			for source in reading:
				while b'\n' in store[source]:
					line,_ = store[source].split(b'\n',1)
					# sys.stderr.write(str(line).replace('\n','\\n') + '\n')
					# sys.stderr.flush()
					sent = write[source](line + b'\n')
					# sys.stderr.write('sent %d\n' % sent)
					# sys.stderr.flush()
					if sent:
						store[source] = store[source][sent:]
						continue
					break
				if backlog[source]:
					store[source] += backlog[source].popleft()

	def run (self):
		if not self.init():
			sys.exit(1)
		try:
			self.loop()
		except KeyboardInterrupt:
			self.cleanup()
			sys.exit(0)
		except Exception as exc:
			sys.stderr.write(str(exc))
			sys.stderr.write('\n\n')
			sys.stderr.flush()
			traceback.print_exc(file=sys.stderr)
			sys.stderr.flush()
			self.cleanup()
			sys.exit(1)


def main (location=''):
	if not location:
		location = os.environ.get('exabgp_cli_pipe','')
	if not location:
		sys.stderr.write("usage %s %s\n" % (sys.executable,' '.join(sys.argv)))
		sys.stderr.write("run with 'env exabgp_cli_pipe=<location>' if you are trying to mess with ExaBGP's intenals")
		sys.stderr.flush()
		sys.exit(1)
	Control(location).run()


if __name__ == '__main__':
	main()



# encoding: utf-8
"""
originatorid.py

Created by Thomas Mangin on 2012-07-07.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.ip import IPv4

from exabgp.bgp.message.update.attribute.attribute import Attribute


# ============================================================== OriginatorID (3)

@Attribute.register()
class OriginatorID (Attribute,IPv4):
	ID = Attribute.CODE.ORIGINATOR_ID
	FLAG = Attribute.Flag.OPTIONAL
	CACHING = True

	__slots__ = []

	def __eq__ (self, other):
		return \
			self.ID == other.ID and \
			self.FLAG == other.FLAG

	def __ne__ (self, other):
		return not self.__eq__(other)

	def pack (self, negotiated=None):
		return self._attribute(self.ton())

	@classmethod
	def unpack (cls, data, negotiated):
		return IPv4.unpack(data,cls)



# encoding: utf-8
"""
ifaceaddr.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import unpack

from exabgp.protocol.ip import IP
from exabgp.util import ordinal

#   https://tools.ietf.org/html/rfc5305#section-3.2
#   This sub-TLV contains a 4-octet IPv4 address for the interface
#   described by the (main) TLV.  This sub-TLV can occur multiple times.
#
#
#   https://tools.ietf.org/html/rfc6119 4.2
#    The IPv6 Interface Address sub-TLV of the Extended IS Reachability
#   TLV has sub-TLV type 12.  It contains a 16-octet IPv6 address for the
#   interface described by the containing Extended IS Reachability TLV.
#   This sub-TLV can occur multiple times.
# ================================================================== InterfaceAddress


class IfaceAddr (object):

	def __init__ (self, iface_addr, packed=None):
		self.iface_address = iface_addr
		self._packed = packed

	@classmethod
	def unpack (cls, data):
		if len(data) == 4:
			# IPv4 address
			addr = IP.unpack(data[:4])
		elif len(data) == 16:
			# IPv6
			addr = IP.unpack(data[:16])
		return cls(iface_addr=addr)

	def json (self, compact=None):
		return '"interface-address": "%s"' % self.iface_address

	def __eq__ (self, other):
		return self.iface_address == other.iface_address

	def __neq__ (self, other):
		return self.iface_address != other.iface_address

	def __lt__ (self, other):
		raise RuntimeError('Not implemented')

	def __le__ (self, other):
		raise RuntimeError('Not implemented')

	def __gt__ (self, other):
		raise RuntimeError('Not implemented')

	def __ge__ (self, other):
		raise RuntimeError('Not implemented')

	def __str__ (self):
		return ':'.join('%02X' % ordinal(_) for _ in self._packed)

	def __repr__ (self):
		return self.__str__()

	def __len__ (self):
		return len(self._packed)

	def __hash__ (self):
		return hash(str(self))

	def pack (self):
		return self._packed



# encoding: utf-8
"""
iso

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
"""


# =========================================================================== ISO
#

from exabgp.vendoring.bitstring import BitArray

class ISO (object):

	def __init__ (self, sysid,
                  selector=None,
                  area_id=None,
                  afi=49):

		self.sysid = sysid
		self.area_id = area_id
		self.selector = selector
		self.afi = afi

	@classmethod
	def unpack_sysid(cls,data):
		b = BitArray(bytes=data)
		return b.hex

	def json(self,compact=None):
		return self.sysid



# encoding: utf-8
"""
community.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import ordinal
from exabgp.bgp.message.update.attribute import Attribute

from struct import pack

# ======================================================= ExtendedCommunity (16)
# XXX: Should subclasses register with transitivity ?


class ExtendedCommunity (Attribute):
	ID = Attribute.CODE.EXTENDED_COMMUNITY
	FLAG = Attribute.Flag.TRANSITIVE | Attribute.Flag.OPTIONAL

	COMMUNITY_TYPE    = 0x00  # MUST be redefined by subclasses
	COMMUNITY_SUBTYPE = 0x00  # MUST be redefined by subclasses
	NON_TRANSITIVE    = 0x40

	registered_extended = {}

	@classmethod
	def register (cls, klass):
		cls.registered_extended[(klass.COMMUNITY_TYPE & 0x0F,klass.COMMUNITY_SUBTYPE)] = klass
		return klass

	# size of value for data (boolean: is extended)
	length_value = {False:7, True:6}
	name = {False: 'regular', True: 'extended'}

	__slots__ = ['community']

	def __init__ (self, community):
		# Two top bits are iana and transitive bits
		self.community = community
		self.klass = None

	def __eq__(self, other):
		return \
			self.ID == other.ID and \
			self.FLAG == other.FLAG and \
			self.community == other.community

	def __ne__(self, other):
		return not self.__eq__(other)

	def __lt__ (self, other):
		return self.community < other.community

	def __le__ (self, other):
		return self.community <= other.community

	def __gt__ (self, other):
		return self.community > other.community

	def __ge__ (self, other):
		return self.community >= other.community

	def iana (self):
		return not not (self.community[0] & 0x80)

	def transitive (self):
		# bit set means "not transitive"
		# RFC4360:
		#   T - Transitive bit
		#     Value 0: The community is transitive across ASes
		#     Value 1: The community is non-transitive across ASes
		return not (self.community[0] & 0x40)

	def pack (self, negotiated=None):
		return self.community

	def _subtype (self, transitive=True):
		# if not transitive -> set the 'transitive' bit, as per RFC4360
		return pack(
			'!BB',
			self.COMMUNITY_TYPE if transitive else self.COMMUNITY_TYPE | self.NON_TRANSITIVE,
			self.COMMUNITY_SUBTYPE
		)

	def json (self):
		h = 0x00
		for byte in self.community:
			h <<= 8
			h += ordinal(byte)
		s = self.klass.__repr__(self) if self.klass else ''
		return '{ "value": %s, "string": "%s" }' % (h,s)

	def __repr__ (self):
		if self.klass:
			return self.klass.__repr__(self)
		h = 0x00
		for byte in self.community:
			h <<= 8
			h += ordinal(byte)
		return "0x%016X" % h

	def __len__ (self):
		return 8

	def __hash__ (self):
		return hash(self.community)

	@staticmethod
	def unpack (data, negotiated=None):
		# 30/02/12 Quagga communities for soo and rt are not transitive when 4360 says they must be, hence the & 0x0FFF
		community = (ordinal(data[0]) & 0x0F,ordinal(data[1]))
		if community in ExtendedCommunity.registered_extended:
			klass = ExtendedCommunity.registered_extended[community]
			instance = klass.unpack(data)
			instance.klass = klass
			return instance
		return ExtendedCommunity(data)



# encoding: utf-8
"""
ospfaddr.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from exabgp.protocol.ip import IP

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //                Forwarding Address (variable)                //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     https://tools.ietf.org/html/rfc7752#section-3.3.3.5


@LINKSTATE.register()
class OspfForwardingAddress(object):
	TLV = 1156

	def __init__ (self, addr):
		self.addr = addr

	def __repr__ (self):
		return "Ospf forwarding address: '%s'" % (self.addr)

	@classmethod
	def unpack (cls,data,length):
		if len(data) == 4:
			# IPv4 address
			addr = IP.unpack(data[:4])
		elif len(data) == 16:
			# IPv6
			addr = IP.unpack(data[:16])
		else:
			raise Notify(3,5, "Error parsing OSPF Forwarding Address. Wrong size")
		return cls(addr=addr)

	def json (self,compact=None):
		return '"ospf-forwarding-address": "%s"' % str(self.addr)



# encoding: utf-8
"""
esi.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2014-2017 Orange. All rights reserved.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes_i

# TODO: take into account E-VPN specs that specify the role of the first bit of ESI
# (since draft-ietf-l2vpn-evpn-05)


# Ethernet Segment Identifier
class ESI (object):
	DEFAULT = concat_bytes_i(character(0) for _ in range(0,10))
	MAX = concat_bytes_i(character(0xFF) for _ in range(0,10))

	__slots__ = ['esi']

	def __init__ (self, esi=None):
		self.esi = self.DEFAULT if esi is None else esi
		if len(self.esi) != 10:
			raise Exception("incorrect ESI, len %d instead of 10" % len(esi))

	def __eq__ (self, other):
		return self.esi == other.esi

	def __neq__ (self, other):
		return self.esi != other.esi

	def __lt__ (self, other):
		raise RuntimeError('comparing ESI for ordering does not make sense')

	def __le__ (self, other):
		raise RuntimeError('comparing ESI for ordering does not make sense')

	def __gt__ (self, other):
		raise RuntimeError('comparing ESI for ordering does not make sense')

	def __ge__ (self, other):
		raise RuntimeError('comparing ESI for ordering does not make sense')

	def __str__ (self):
		if self.esi == self.DEFAULT:
			return "-"
		return ":".join('%02x' % ordinal(_) for _ in self.esi)

	def __repr__ (self):
		return self.__str__()

	def pack (self):
		return self.esi

	def __len__ (self):
		return 10

	def __hash__ (self):
		return hash(self.esi)

	@classmethod
	def unpack (cls, data):
		return cls(data[:10])

	def json (self, compact=None):
		return '"esi": "%s"' % str(self)



# encoding: utf-8
"""
labels.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.util import character
from exabgp.util import concat_bytes_i


# ======================================================================= Labels
# RFC 3107

def opt_raw_label(label, format=' (%d)'):
    return format % label if label else ''

class Labels (object):
	MAX = pow(2,20)-1

	__slots__ = ['labels','packed','_len','raw_labels']

	def __init__ (self, labels, bos=True, raw_labels=None):
		self.labels = labels
		self.raw_labels = raw_labels
		packed = []
		if raw_labels:
			for label in raw_labels:
				packed.append(pack('!L',label)[1:])
			# fill self.labels as well, not for packing, but to allow
			# consistent string representations
			if not self.labels:
				self.labels = [x >> 4 for x in self.raw_labels]
		else:
			for label in labels:
				# shift to 20 bits of the label to be at the top of three bytes and then truncate.
				packed.append(pack('!L',label << 4)[1:])
			# Mark the bottom of stack with the bit
			if packed and bos:
				packed.pop()
				packed.append(pack('!L',(label << 4) | 1)[1:])
			self.raw_labels = [None for _ in self.labels]
		self.packed = concat_bytes_i(packed)
		self._len = len(self.packed)

	def __eq__ (self, other):
		return self.labels == other.labels

	def __neq__ (self, other):
		return self.labels != other.labels

	def __lt__ (self, other):
		raise RuntimeError('comparing EthernetTag for ordering does not make sense')

	def __le__ (self, other):
		raise RuntimeError('comparing EthernetTag for ordering does not make sense')

	def __gt__ (self, other):
		raise RuntimeError('comparing EthernetTag for ordering does not make sense')

	def __ge__ (self, other):
		raise RuntimeError('comparing EthernetTag for ordering does not make sense')

	def pack (self):
		return self.packed

	def __len__ (self):
		return self._len

	def json (self):
		if len(self.labels) >= 1:
			return '"label": [ %s ]' % ', '.join(["[%d%s]" % (l,opt_raw_label(r, ', %d')) for (l,r) in zip(self.labels, self.raw_labels)])
		else:
			return ''

	def __str__ (self):
		if len(self.labels) > 1:
			return ' label [ %s ]' % ' '.join(["%d%s" % (l,opt_raw_label(r)) for (l,r) in zip(self.labels, self.raw_labels)])
		elif len(self.labels) == 1:
			return ' label %d%s' % (self.labels[0],opt_raw_label(self.raw_labels[0]))
		else:
			return ''

	def __repr__(self):
		if len(self.labels) > 1:
			return '[ %s ]' % ','.join(["%d%s" % (l,opt_raw_label(r)) for (l,r) in zip(self.labels, self.raw_labels)])
		elif len(self.labels) == 1:
			return '%d%s' % (self.labels[0],opt_raw_label(self.raw_labels[0]))
		else:
			return '[ ]'

	@classmethod
	def unpack (cls, data):
		labels = []
		raw_labels = []
		while len(data):
			label = unpack('!L',character(0)+data[:3])[0]
			data = data[3:]
			labels.append(label >> 4)
			raw_labels.append(label)
			if label & 0x001:
				break
		return cls(labels, raw_labels=raw_labels)


Labels.NOLABEL = Labels([])



# encoding: utf-8
"""
line/reactor.py

Created by Thomas Mangin on 2017-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.version import version as _version
from exabgp.reactor.api.command.command import Command


def register_reactor ():
	pass


@Command.register('text','help',False)
def manual (self, reactor, service, _):
	lines = []
	for command in sorted(self.callback['text']):
		if self.callback['options'][command]:
			extended = '%s [ %s ]' % (command, ' | '.join(self.callback['options'][command]))
		else:
			extended = command
		lines.append('[neighbor <ip> [filters]] ' + command if self.callback['neighbor'][command] else '%s ' % extended)

	reactor.processes.answer(service,'',True)
	reactor.processes.answer(service,'available API commands are listed here:',True)
	reactor.processes.answer(service,'=======================================',True)
	reactor.processes.answer(service,'',True)
	reactor.processes.answer(service,'filter can be: [local-ip <ip>][local-as <asn>][peer-as <asn>][router-id <router-id>]',True)
	reactor.processes.answer(service,'',True)
	reactor.processes.answer(service,'command are:',True)
	reactor.processes.answer(service,'------------',True)
	reactor.processes.answer(service,'',True)
	for line in sorted(lines):
		reactor.processes.answer(service,line,True)
	reactor.processes.answer(service,'',True)
	reactor.processes.answer_done(service)
	return True


@Command.register('text','shutdown',False)
def shutdown (self, reactor, service, _):
	reactor.signal.received = reactor.signal.SHUTDOWN
	reactor.processes.answer(service,'shutdown in progress')
	reactor.processes.answer_done(service)
	return True


@Command.register('text','reload',False)
def reload (self, reactor, service, _):
	reactor.signal.received = reactor.signal.RELOAD
	reactor.processes.answer(service,'reload in progress')
	reactor.processes.answer_done(service)
	return True


@Command.register('text','restart',False)
def restart (self, reactor, service, _):
	reactor.signal.received = reactor.signal.RESTART
	reactor.processes.answer(service,'restart in progress')
	reactor.processes.answer_done(service)
	return True


@Command.register('text','version',False)
def version (self, reactor, service, _):
	reactor.processes.answer(service,'exabgp %s' % _version,force=True)
	reactor.processes.answer_done(service)
	return True


@Command.register('text','#',False)
def comment (self, reactor, service, line):
	self.logger.debug(line.lstrip().lstrip('#').strip(),'process')
	reactor.processes.answer_done(service)
	return True


@Command.register('text','reset',False)
def reset (self, reactor, service, line):
	reactor.asynchronous.clear(service)


@Command.register('text','crash')
def crash (self, reactor, service, line):
	def callback():
		raise ValueError('crash test of the API')
		yield None
	reactor.asynchronous.schedule(service, line, callback())
	return True



#!/usr/bin/env python
# encoding: utf-8
"""
open.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import sys
import unittest

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message import Message
from exabgp.bgp.message import Open
from exabgp.bgp.message.open import Version
from exabgp.bgp.message.open import ASN
from exabgp.bgp.message.open import RouterID
from exabgp.bgp.message.open import HoldTime
from exabgp.bgp.message.open.routerid import RouterID
from exabgp.bgp.message.open.capability import Capabilities
from exabgp.bgp.message.open.capability import RouteRefresh

from exabgp.util.test import data_from_body

from exabgp.configuration.environment import environment
environment.setup('')


open_body = [0x4, 0xff, 0xfe, 0x0, 0xb4, 0x0, 0x0, 0x0, 0x0, 0x20, 0x2, 0x6, 0x1, 0x4, 0x0, 0x1, 0x0, 0x1, 0x2, 0x6, 0x1, 0x4, 0x0, 0x2, 0x0, 0x1, 0x2, 0x2, 0x80, 0x0, 0x2, 0x2, 0x2, 0x0, 0x2, 0x6, 0x41, 0x4, 0x0, 0x0, 0xff, 0xfe]

class TestData (unittest.TestCase):

	def test_1_open (self):
		check_capa = {
			1: [(AFI.ipv4, SAFI.unicast), (AFI.ipv6, SAFI.unicast)],
			2: RouteRefresh(),
			65: 65534,
			128: RouteRefresh(),
		}

		message_id = 1
		negotiated = {'invalid':'test'}

		o = Message.unpack(message_id,data_from_body(open_body),negotiated)

		self.assertEqual(o.version,4)
		self.assertEqual(o.asn,65534)
		self.assertEqual(o.router_id,RouterID('0.0.0.0'))
		self.assertEqual(o.hold_time,180)
		for k,v in o.capabilities.items():
			self.assertEqual(v,check_capa[k])

	def test_2_open (self):
		capabilities = Capabilities()
		o = Open(Version(4),ASN(65500),HoldTime(180),RouterID('127.0.0.1'),capabilities)
		self.assertEqual(o.version,4)
		self.assertEqual(o.asn,65500)
		self.assertEqual(o.router_id,RouterID('127.0.0.1'))
		self.assertEqual(o.hold_time,180)
		self.assertEqual(o.capabilities, {})

if __name__ == '__main__':
	unittest.main()



#!/usr/bin/env python

import os
import sys
import time

# When the parent dies we are seeing continual newlines, so we only access so many before stopping
counter = 0

while True:
	try:
		line = sys.stdin.readline().strip()
		if line == "":
			counter += 1
			if counter > 100:
				break
			continue

		counter = 0

		send = '\n%s %s %s\n' % ( '-'*10, line, '-'*10 )
		print >> sys.stderr, send
		sys.stderr.flush()
	except KeyboardInterrupt:
		pass
	except IOError:
		# most likely a signal during readline
		pass



# encoding: utf-8
"""
logger.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import print_function

import os
import sys
import stat
import time
import syslog
import logging
import logging.handlers
import pdb
from collections import deque

from exabgp.util.od import od
from exabgp.configuration.environment import environment


def _can_write (location):
	try:
		s  = os.stat(os.path.dirname(location))
	except OSError:
		return None
	mode = s[stat.ST_MODE]
	uid  = os.geteuid()
	gid  = os.getegid()

	return not not (
		((s[stat.ST_UID] == uid) and (mode & stat.S_IWUSR)) or
		((s[stat.ST_GID] == gid) and (mode & stat.S_IWGRP)) or
		(mode & stat.S_IWOTH)
	)


# This delays the evaluation of the od() function which is expensive
# hence why pylint too-few-public-method is disabled

class LazyFormat (object):
	def __init__ (self, prefix, message, formater=od):
		self.prefix = prefix
		self.message = message
		self.formater = formater

	def split (self, char):
		return str(self).split(char)

	def __str__ (self):
		formated = self.formater(self.message)
		return '%s (%4d) %s' % (self.prefix,len(self.message),formated)


class LazyAttribute (object):
	def __init__ (self, flag, aid, length, data):
		self.flag = flag
		self.aid = aid
		self.length = length
		self.data = data

	def split (self, char):
		return str(self).split(char)

	def __str__ (self):
		return 'attribute %-18s flag 0x%02x type 0x%02x len 0x%02x%s' % (
			str(self.aid),
			self.flag,
			int(self.aid),
			self.length,
			' payload %s' % od(self.data) if self.data else ''
		)


class LazyNLRI (object):
	def __init__ (self, afi, safi, addpath, data):
		self.afi = afi
		self.safi = safi
		self.addpath = addpath
		self.data = data

	def split (self, char):
		return str(self).split(char)

	def __str__ (self):
		family = '%s %s' % (self.afi,self.safi)
		path = 'with path-information' if self.addpath else 'without path-information'
		payload = od(self.data) if self.data else 'none'
		return 'NLRI      %-18s %-28s payload %s' % (family,path,payload)


def istty (std):
	try:
		return std.isatty()
	except KeyboardInterrupt:
		raise
	except Exception:
		return False


class Logger (object):
	RECORD = {
		'START':    '\033[01;32m',  # Green
		'DEBUG':    '',
		'INFO':     '\033[01;32m',  # Green
		'NOTICE':   '\033[01;34m',  # Blue
		'WARNING':  '\033[01;33m',  # Yellow
		'ERR':      '\033[01;31m',  # Red
		'CRIT':     '\033[00;31m',  # Strong Red
	}

	MESSAGE = {
		'START':    '\033[1m',  # Green
		'DEBUG':    '',
		'INFO':     '\033[1m',  # Green
		'NOTICE':   '\033[1m',  # Blue
		'WARNING':  '\033[1m',  # Yellow
		'ERR':      '\033[1m',  # Red
		'CRIT':     '\033[1m',  # Strong Red
	}

	END = '\033[0m'

	TTY = {
		'stderr': lambda: istty(sys.stderr),
		'stdout': lambda: istty(sys.stdout),
		'out': lambda: istty(sys.stdout),
	}

	_instance = dict()
	_syslog = None
	_where = ''

	_history = deque()
	_max_history = 20

	_default = None

	_config = ''
	_pid = os.getpid()
	_cwd = os.getcwd()

	def __new__ (cls):
		if cls._instance.get('class',None) is None:
			return super(Logger, cls).__new__(cls)
		else:
			return cls._instance['class']

	# we use os.pid everytime as we may fork and the class is instance before it

	def pdb (self, level):
		if self._option['pdb'] and level == 'CRIT':
			# not sure why, pylint reports an import error here
			pdb.set_trace()

	def config (self, config=None):
		if config is not None:
			self._config = config
		return self._config

	def history (self):
		return "\n".join(self._format(*_) for _ in self._history)

	def _record (self, timestamp, message, source, level):
		if len(self._history) > self._max_history:
			self._history.popleft()
		self._history.append((message,source,level,timestamp))

	def __init__ (self):
		if self._instance.get('class',None) is not None:
			return

		self._instance['class'] = self

		command = environment.settings()
		self.short = command.log.short
		self.level = command.log.level

		self._option = {
			'pdb':           command.debug.pdb,
			'reactor':       command.log.enable and (command.log.all or command.log.reactor),
			'daemon':        command.log.enable and (command.log.all or command.log.daemon),
			'processes':     command.log.enable and (command.log.all or command.log.processes),
			'configuration': command.log.enable and (command.log.all or command.log.configuration),
			'network':       command.log.enable and (command.log.all or command.log.network),
			'wire':          command.log.enable and (command.log.all or command.log.packets),
			'message':       command.log.enable and (command.log.all or command.log.message),
			'rib':           command.log.enable and (command.log.all or command.log.rib),
			'timer':         command.log.enable and (command.log.all or command.log.timers),
			'routes':        command.log.enable and (command.log.all or command.log.routes),
			'parser':        command.log.enable and (command.log.all or command.log.parser),
		}

		if not command.log.enable:
			self.destination = ''
			return

		self.destination = command.log.destination

		self.restart(True)

	def _local_syslog (self):
		if sys.platform == "darwin":
			address = '/var/run/syslog'
		else:
			address = '/dev/log'
		if not os.path.exists(address):
			address = ('localhost', 514)
		handler = logging.handlers.SysLogHandler(address)

		self._syslog.addHandler(handler)
		return True

	def _remote_syslog (self, destination):
		# If the address is invalid, each syslog call will print an error.
		# See how it can be avoided, as the socket error is encapsulated and not returned
		address = (destination, 514)
		handler = logging.handlers.SysLogHandler(address)

		self._syslog.addHandler(handler)
		return True

	def _standard (self, facility):
		handler = logging.StreamHandler(getattr(sys,facility))

		self._syslog.addHandler(handler)
		return True

	def _file (self, destination):
		# folder
		logfile = os.path.realpath(os.path.normpath(os.path.join(self._cwd,destination)))
		can = _can_write(logfile)
		if can is True:
			handler = logging.handlers.RotatingFileHandler(logfile, maxBytes=5*1024*1024, backupCount=5)
		elif can is None:
			self.critical('ExaBGP can not access (perhaps as it does not exist) the log folder provided','logger')
			return False
		else:
			self.critical('ExaBGP does not have the right to write in the requested log directory','logger')
			return False

		self._syslog.addHandler(handler)
		return True

	def restart (self, first=False):
		try:
			if first:
				self._where = 'stdout'
				self._default = logging.StreamHandler(sys.stdout)
				self._syslog = logging.getLogger()
				self._syslog.setLevel(logging.DEBUG)
				self._syslog.addHandler(self._default)
				return True
		except IOError:
			# no way to report anything via stdout, silently failing
			return False

		if not self._syslog:
			# no way to report anything via stdout, silently failing
			return False

		for handler in self._syslog.handlers:
			self._syslog.removeHandler(handler)
		self._syslog.addHandler(self._default)

		try:
			if self.destination == 'stderr':
				self._where = 'stderr'
				return True
			elif self.destination == 'stdout':
				self._where = 'out'
				result = self._standard(self.destination)
			elif self.destination in ('','syslog'):
				self._where = 'syslog'
				result = self._local_syslog()
			elif self.destination.startswith('host:'):
				self._where = 'syslog'
				result = self._remote_syslog(self.destination[5:].strip())
			else:
				self._where = 'file'
				result = self._file(self.destination)

			if result:
				self._syslog.removeHandler(self._default)
			return result
		except IOError:
			self.critical('Can not set logging (are stdout/stderr closed?)','logger')
			return False

	def _format (self, message, source, level,timestamp=None):
		if timestamp is None:
			timestamp = time.localtime()
			self._record(timestamp,message,source,level)

		if self.short:
			return message

		if self._where in ['stdout','stderr','out']:
			now = time.strftime('%H:%M:%S',timestamp)
			if not self.TTY[self._where]():
				return "%s | %-6d | %-15s | %s" % (now,self._pid,source,message)
			return "%s | %-6d | %s%-13s%s | %s%-8s%s" % (
				now,
				self._pid,
				self.RECORD.get(level,''),source,self.END,
				self.MESSAGE.get(level,''),message,self.END
			)
		elif self._where in ['syslog',]:
			return "%s[%d]: %-13s %s" % (environment.application,self._pid,source,message)
		elif self._where in ['file',]:
			now = time.strftime('%a, %d %b %Y %H:%M:%S',timestamp)
			return "%s %-6d %-13s %s" % (now,self._pid,source,message)
		else:
			# failsafe
			return "%s | %-8s | %-6d | %-13s | %s" % (now,level,self._pid,source,message)

	def _report (self, message, source, level):
		if source.startswith('incoming-'):
			src = 'wire'
		elif source.startswith('outgoing-'):
			src = 'wire'
		elif source.startswith('ka-'):
			src = 'timer'
		elif source.startswith('peer-'):
			src = 'network'
		else:
			src = source

		log = self._option.get(src,True) and getattr(syslog,'LOG_%s' % level) <= self.level

		if not log:
			return

		for line in message.split('\n'):
			if self._syslog:
				self._syslog.debug(self._format(line,source,level))
			else:
				print(self._format(line,source,level))
				sys.stdout.flush()

	def debug (self, message, source='', level='DEBUG'):
		self._report(message,source,level)

	def info (self, message, source='', level='INFO'):
		self._report(message,source,level)

	def notice (self, message, source='', level='NOTICE'):
		self._report(message,source,level)

	def warning (self, message, source='', level='WARNING'):
		self._report(message,source,level)

	def error (self, message, source='', level='ERR'):
		self._report(message,source,level)

	def critical (self, message, source='', level='CRIT'):
		self._report(message,source,level)


class FakeLogger (object):
	def __getattr__ (self, name):
		def printf (data, _=None):
			sys.stdout.write('Fake logger [%s]\n' % str(data))
		return printf



# encoding: utf-8
"""
vpls.py

Created by Nikita Shirokov on 2014-06-16.
Copyright (c) 2014-2017 Nikita Shirokov. All rights reserved.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import unpack
from struct import pack
from exabgp.vendoring import six
from exabgp.util import concat_bytes
from exabgp.protocol.ip import IP
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message.direction import OUT
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher


def _unique ():
	value = 0
	while True:
		yield value
		value += 1


unique = _unique()


@NLRI.register(AFI.l2vpn,SAFI.vpls)
class VPLS (NLRI):

	__slots__ = ['action','nexthop','rd','base','offset','size','endpoint','unique']

	# XXX: Should take AFI, SAFI and OUT.direction as parameter to match other NLRI
	def __init__ (self, rd, endpoint, base, offset, size):
		NLRI.__init__(self,AFI.l2vpn,SAFI.vpls)
		self.action = OUT.ANNOUNCE
		self.nexthop = None
		self.rd = rd
		self.base = base
		self.offset = offset
		self.size = size
		self.endpoint = endpoint
		self.unique = six.next(unique)

	def feedback (self, action):
		if self.nexthop is None and action == OUT.ANNOUNCE:
			return 'vpls nlri next-hop missing'
		if self.endpoint is None:
			return 'vpls nlri endpoint missing'
		if self.base is None:
			return 'vpls nlri base missing'
		if self.offset is None:
			return 'vpls nlri offset missing'
		if self.size is None:
			return 'vpls nlri size missing'
		if self.rd is None:
			return 'vpls nlri route-distinguisher missing'
		if self.base > (0xFFFFF - self.size):  # 20 bits, 3 bytes
			return 'vpls nlri size inconsistency'
		return ''

	def assign (self, name, value):
		setattr(self,name,value)

	def pack_nlri (self, negotiated=None):
		return concat_bytes(
			b'\x00\x11',  # pack('!H',17)
			self.rd.pack(),
			pack(
				'!HHH',
				self.endpoint,
				self.offset,
				self.size
			),
			pack(
				'!L',
				(self.base << 4) | 0x1
			)[1:]  # setting the bottom of stack, should we ?
		)

	# XXX: FIXME: we need an unique key here.
	# XXX: What can we use as unique key ?
	def json (self,compact=None):
		content = ', '.join([
			self.rd.json(),
			'"endpoint": %s' % self.endpoint,
			'"base": %s' % self.base,
			'"offset": %s' % self.offset,
			'"size": %s' % self.size,
		])
		return '{ %s }' % (content)

	def extensive (self):
		return "vpls%s endpoint %s base %s offset %s size %s %s" % (
			self.rd,
			self.endpoint,
			self.base,
			self.offset,
			self.size,
			'' if self.nexthop is None else 'next-hop %s' % self.nexthop,
		)

	def __str__ (self):
		return self.extensive()

	@classmethod
	def unpack_nlri (cls, afi, safi, bgp, action, addpath):
		# label is 20bits, stored using 3 bytes, 24 bits
		length, = unpack('!H',bgp[0:2])
		if len(bgp) != length+2:
			raise Notify(3,10,'l2vpn vpls message length is not consistent with encoded bgp')
		rd = RouteDistinguisher(bgp[2:10])
		endpoint,offset,size = unpack('!HHH',bgp[10:16])
		base = unpack('!L',b'\x00'+bgp[16:19])[0] >> 4
		nlri = cls(rd,endpoint,base,offset,size)
		nlri.action = action
		# nlri.nexthop = IP.unpack(nexthop)
		return nlri, bgp[19:]



#!/usr/bin/env python

from __future__ import print_function
import os
import sys
import time

flow="flow route { match { source 10.0.0.1/32; destination 10.0.0.2/32; destination-port =3128; protocol tcp; } then { discard; } }"

# When the parent dies we are seeing continual newlines, so we only access so many before stopping
counter = 1

# sleep a little bit or we will never see the asm in the configuration file
# and the message received just before we go to the established loop will be printed twice
time.sleep(1)

while True:
	try:
		time.sleep(1)
		if counter % 2:
			print('announce', flow)
			sys.stdout.flush()
		else:
			print('withdraw', flow)
			sys.stdout.flush()

		counter += 1
	except KeyboardInterrupt:
		pass
	except IOError:
		break



# encoding: utf-8
"""
store.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.bgp.message import IN
from exabgp.bgp.message import OUT


class Cache (object):
	def __init__ (self, cache, families):
		self.cache = cache
		self._seen = {}
		# self._seen[family][change-index] = change
		# nlri.index() would be a few bytes shorter than change.index() but ..
		# we need change.index() in other part of the code
		# we pre-compute change.index() so that it is only allocted once
		self.families = families

	def clear_cache (self):
		self._seen = {}

	def delete_cached_family (self, families):
		for family in self._seen.keys():
			if family not in families:
				del self._seen[family]

	def cached_changes (self, families=None):
		# families can be None or []
		requested_families = self.families if families is None else set(families).intersection(self.families)

		# we use list() to make a snapshot of the data at the time we run the command
		for family in requested_families:
			for change in self._seen.get(family,{}).values():
				if change.nlri.action == OUT.ANNOUNCE:
					yield change

	def is_cached (self,change):
		if not self.cache:
			return False
		# if we cache sent NLRI and this NLRI was never sent before, we do not need to send a withdrawal
		# as the route removed before we could announce it
		return change.index() not in self._seen.get(change.nlri.family(),{})

	# return a tuple exists in cache, same in cache
	def in_cache (self,change):
		if not self.cache:
			return False,False

		old_change = self._seen.get(change.nlri.family(),{}).get(change.index(),None)
		if not old_change:
			return False,False

		if change.nlri.action != OUT.ANNOUNCE:
			return True,False

		if old_change.attributes.index() != change.attributes.index():
			return True,False

		if old_change.nlri.nexthop.index() != change.nlri.nexthop.index():
			return True,False

		return True,True

	# add a change to the cache of seen Change
	def update_cache (self,change):
		if not self.cache:
			return
		family = change.nlri.family()
		index = change.index()
		if change.nlri.action == OUT.ANNOUNCE:
			self._seen.setdefault(family,{})[index] = change
		elif family in self._seen:
			self._seen[family].pop(index,None)



# encoding: utf-8
"""
command/limit.py

Created by Thomas Mangin on 2017-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import re


def extract_neighbors (command):
	"""Return a list of neighbor definition : the neighbor definition is a list of string which are in the neighbor indexing string"""
	# This function returns a list and a string
	# The first list contains parsed neighbor to match against our defined peers
	# The string is the command to be run for those peers
	# The parsed neighbor is a list of the element making the neighbor string so each part can be checked against the neighbor name

	returned = []
	neighbor,remaining = command.split(' ',1)
	if neighbor != 'neighbor':
		return [],command

	ip,command = remaining.split(' ',1)
	definition = ['neighbor %s' % (ip)]

	while True:
		try:
			key,value,remaining = command.split(' ',2)
		except ValueError:
			key,value = command.split(' ',1)
		if key == ',':
			returned.append(definition)
			_,command = command.split(' ',1)
			definition = []
			continue
		if key not in ['neighbor','local-ip','local-as','peer-as','router-id','family-allowed']:
			if definition:
				returned.append(definition)
			break
		definition.append('%s %s' % (key,value))
		command = remaining

	return returned,command


def match_neighbor (description, name):
	for string in description:
		if re.search(r'(^|\s)%s($|\s|,)' % re.escape(string), name) is None:
			return False
	return True


def match_neighbors (peers,descriptions):
	"""Return the sublist of peers matching the description passed, or None if no description is given"""
	if not descriptions:
		return peers.keys()

	returned = []
	for key in peers:
		for description in descriptions:
			if match_neighbor(description,key):
				if key not in returned:
					returned.append(key)
	return returned



# encoding: utf-8
"""
operational/__init__.py

Created by Thomas Mangin on 2015-06-23.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import sys

from exabgp.util.ip import isipv4

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.bgp.message.open.routerid import RouterID
from exabgp.bgp.message.operational import MAX_ADVISORY
from exabgp.bgp.message.operational import Advisory
from exabgp.bgp.message.operational import Query
from exabgp.bgp.message.operational import Response


if sys.version_info > (3,):
	long = int


def _operational (klass, parameters, tokeniser):
	def utf8 (string):
		return string.encode('utf-8')

	def valid (_):
		return True

	def u32 (_):
		return int(_) <= 0xFFFFFFFF

	def u64 (_):
		return long(_) <= 0xFFFFFFFFFFFFFFFF

	def advisory (_):
		return len(_.encode('utf-8')) <= MAX_ADVISORY + 2  # the two quotes

	convert = {
		'afi': AFI.value,
		'safi': SAFI.value,
		'sequence': int,
		'counter': long,
		'advisory': utf8
	}

	validate = {
		'afi': AFI.value,
		'safi': SAFI.value,
		'sequence': u32,
		'counter': u64,
	}

	number = len(parameters)*2
	tokens = []
	while len(tokens) != number:
		tokens.append(tokeniser())

	data = {}

	while tokens and parameters:
		command = tokens.pop(0).lower()
		value = tokens.pop(0)

		if command == 'router-id':
			if isipv4(value):
				data['routerid'] = RouterID(value)
			else:
				raise ValueError('invalid operational value for %s' % command)
			continue

		expected = parameters.pop(0)

		if command != expected:
			raise ValueError('invalid operational syntax, unknown argument %s' % command)
		if not validate.get(command,valid)(value):
			raise ValueError('invalid operational value for %s' % command)

		data[command] = convert[command](value)

	if tokens or parameters:
		raise ValueError('invalid advisory syntax, missing argument(s) %s' % ', '.join(parameters))

	if 'routerid' not in data:
		data['routerid'] = None

	return klass(**data)


_dispatch = {}


def register (name):
	def inner (function):
		_dispatch[name] = function
		return function
	return inner


@register('asm')
def asm (tokeniser):
	return _operational(Advisory.ASM,['afi','safi','advisory'],tokeniser)


@register('adm')
def adm (tokeniser):
	return _operational(Advisory.ADM,['afi','safi','advisory'],tokeniser)


@register('rpcq')
def rpcq (tokeniser):
	return _operational(Query.RPCQ,['afi','safi','sequence'],tokeniser)


@register('rpcp')
def rpcp (tokeniser):
	return _operational(Response.RPCP,['afi','safi','sequence','counter'],tokeniser)


@register('apcq')
def apcq (tokeniser):
	return _operational(Query.APCQ,['afi','safi','sequence'],tokeniser)


@register('apcp')
def apcp (tokeniser):
	return _operational(Response.APCP,['afi','safi','sequence','counter'],tokeniser)


@register('lpcq')
def lpcq (tokeniser):
	return _operational(Query.LPCQ,['afi','safi','sequence'],tokeniser)


@register('lpcp')
def lpcp (tokeniser):
	return _operational(Response.LPCP,['afi','safi','sequence','counter'],tokeniser)


def operational (what, tokeniser):
	return _dispatch.get(what,lambda _: None)(tokeniser)



#!/usr/bin/env python
# encoding: utf-8
"""
connection.py

Created by Thomas Mangin on 2013-07-13.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os
import sys
import cProfile

import unittest

from exabgp.util.od import od
from exabgp.vendoring import six

def test ():
	OPEN = ''.join([chr(int(_,16)) for _ in "FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 1D 01 04 78 14 00 5A 52 DB 00 45 00".split()])
	KEEP = ''.join([chr(int(_,16)) for _ in "FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 00 04".split()])

	from exabgp.reactor.network.outgoing import Outgoing
	connection = Outgoing(1,'82.219.0.69','82.219.212.34')
	writer = connection.writer(OPEN)
	while six.next(writer) is False:
		pass
	writer = connection.writer(KEEP)
	while six.next(writer) is False:
		pass

	reader = connection.reader()

	for size,msg,header,body,notification in reader:
		if size:
			print(od(header+body))
		else:
			sys.stdout.write('-')

	reader = connection.reader()

	for size,msg,header,body,notification in reader:
		if size:
			print(od(header+body))
		else:
			sys.stdout.write('+')

	connection.close()


class TestData (unittest.TestCase):

	def test_1 (self):
		# if not os.environ.get('profile',False):
		# 	result = test()
		# 	if result:
		# 		self.fail(result)
		pass

	def test_2 (self):
		# if not not os.environ.get('profile',False):
		# 	cProfile.run('test()')
		pass

if __name__ == '__main__':
	unittest.main()

	# import cProfile
	# print 'profiling'
	# cProfile.run('unittest.main()','profile.info')



# encoding: utf-8
"""
then.py

Created by Thomas Mangin on 2015-06-22.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.core import Section

from exabgp.configuration.flow.parser import accept
from exabgp.configuration.flow.parser import discard
from exabgp.configuration.flow.parser import rate_limit
from exabgp.configuration.flow.parser import redirect
from exabgp.configuration.flow.parser import redirect_next_hop
from exabgp.configuration.flow.parser import copy
from exabgp.configuration.flow.parser import mark
from exabgp.configuration.flow.parser import action

from exabgp.configuration.static.parser import community
from exabgp.configuration.static.parser import large_community
from exabgp.configuration.static.parser import extended_community


class ParseFlowThen (Section):
	definition = [
		'accept',
		'discard',
		'rate-limit 9600',
		'redirect 30740:12345',
		'redirect 1.2.3.4:5678',
		'redirect 1.2.3.4',
		'redirect-next-hop',
		'copy 1.2.3.4',
		'mark 123',
		'action sample|terminal|sample-terminal',
	]

	syntax = \
		'then {\n' \
		'  %s;\n' \
		'}' % ';\n  '.join(definition)

	known = {
		'accept':              accept,
		'discard':             discard,
		'rate-limit':          rate_limit,
		'redirect':            redirect,
		'redirect-to-nexthop': redirect_next_hop,
		'copy':                copy,
		'mark':                mark,
		'action':              action,
		'community':           community,
		'large-community':     large_community,
		'extended-community':  extended_community,
	}

	# 'community','extended-community'

	action = {
		'accept':              'nop',
		'discard':             'attribute-add',
		'rate-limit':          'attribute-add',
		'redirect':            'nexthop-and-attribute',
		'redirect-to-nexthop': 'attribute-add',
		'copy':                'nexthop-and-attribute',
		'mark':                'attribute-add',
		'action':              'attribute-add',
		'community':           'attribute-add',
		'large-community':     'attribute-add',
		'extended-community':  'attribute-add',
	}

	name = 'flow/then'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		pass

	def pre (self):
		return True

	def post (self):
		return True



# encoding: utf-8
"""
peer.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import time
from collections import defaultdict

# import traceback
from exabgp.vendoring import six
from exabgp.util import ordinal
from exabgp.bgp.timer import ReceiveTimer
from exabgp.bgp.message import Message
from exabgp.bgp.fsm import FSM
from exabgp.bgp.message.open.capability import Capability
from exabgp.bgp.message.open.capability import REFRESH
from exabgp.bgp.message import NOP
from exabgp.bgp.message import Update
from exabgp.bgp.message.refresh import RouteRefresh
from exabgp.bgp.message import Notification
from exabgp.bgp.message import Notify
from exabgp.reactor.protocol import Protocol
from exabgp.reactor.delay import Delay
from exabgp.reactor.keepalive import KA
from exabgp.reactor.network.error import NetworkError
from exabgp.reactor.api.processes import ProcessError

from exabgp.rib.change import Change

from exabgp.configuration.environment import environment
from exabgp.logger import Logger
from exabgp.logger import FakeLogger
from exabgp.logger import LazyFormat

from exabgp.util.trace import trace

from exabgp.util.panic import NO_PANIC
from exabgp.util.panic import FOOTER


class ACTION (object):
	CLOSE = 0x01  # finished, no need to restart the peer
	LATER = 0x02  # re-run at the next reactor round
	NOW   = 0x03  # re-run immediatlely
	ALL   = [CLOSE, LATER, NOW]


class SEND (object):
	DONE    = 0x01
	NORMAL  = 0x02
	REFRESH = 0x04
	ALL     = [DONE, NORMAL, REFRESH]


# As we can not know if this is our first start or not, this flag is used to
# always make the program act like it was recovering from a failure
# If set to FALSE, no EOR and OPEN Flags set for Restart will be set in the
# OPEN Graceful Restart Capability
FORCE_GRACEFUL = True


class Interrupted (Exception):
	pass


class Stop (Exception):
	pass

# ======================================================================== Peer
# Present a File like interface to socket.socket

class Peer (object):
	def __init__ (self, neighbor, reactor):
		try:
			self.logger = Logger()
			# We only to try to connect via TCP once
			self.once = environment.settings().tcp.once
			self.bind = True if environment.settings().tcp.bind else False
		except RuntimeError:
			self.logger = FakeLogger()
			self.once = False
			self.bind = True

		now = time.time()

		self.reactor = reactor
		self.neighbor = neighbor
		# The next restart neighbor definition
		self._neighbor = None

		self.proto = None
		self.fsm = FSM(self,FSM.IDLE)
		self.stats = {
			'fsm':      self.fsm,
			'creation': now,
			'reset': now,
			'complete': 0,
		}
		self.generator = None

		# The peer should restart after a stop
		self._restart = True
		# The peer was restarted (to know what kind of open to send for graceful restart)
		self._restarted = FORCE_GRACEFUL

		# We want to remove routes which are not in the configuration anymore after a signal to reload
		self._reconfigure = True
		# We want to send all the known routes
		self._resend_routes = SEND.DONE

		# We have been asked to teardown the session with this code
		self._teardown = None

		self._delay = Delay()
		self.recv_timer = None

	def id (self):
		return 'peer-%s' % self.neighbor.uid

	def _reset (self, message='',error=''):
		self.fsm.change(FSM.IDLE)
		self.stats = {
			'fsm':      self.fsm,
			'creation': self.stats['creation'],
			'reset':    time.time(),
			'complete': 0,
		}
		if self.proto:
			self.proto.close(u"peer reset, message [{0}] error[{1}]".format(message, error))
		self._delay.increase()

		self.proto = None

		if not self._restart or self.neighbor.generated:
			self.generator = False
			return

		self.generator = None
		self._teardown = None
		self.neighbor.rib.reset()

		# If we are restarting, and the neighbor definition is different, update the neighbor
		if self._neighbor:
			self.neighbor = self._neighbor
			self._neighbor = None

	def _stop (self, message):
		self.generator = False
		if self.proto:
			self.proto.close('stop, message [%s]' % message)
			self.proto = None

	# logging

	def me (self, message):
		return "peer %s ASN %-7s %s" % (self.neighbor.peer_address,self.neighbor.peer_as,message)

	# control

	def stop (self):
		self._teardown = 3
		self._restart = False
		self._restarted = False
		self._delay.reset()
		self.fsm.change(FSM.IDLE)
		self.stats = {
			'fsm':      self.fsm,
			'creation': self.stats['creation'],
			'reset':    time.time(),
		}
		self.neighbor.rib.uncache()

	def remove (self):
		self._stop("removed")
		self.stop()

	def shutdown (self):
		self._stop("shutting down")
		self.stop()

	def resend (self):
		self._resend_routes = SEND.NORMAL
		self._delay.reset()

	def reestablish (self, restart_neighbor=None):
		# we want to tear down the session and re-establish it
		self._teardown = 3
		self._restart = True
		self._restarted = True
		self._resend_routes = SEND.NORMAL
		self._neighbor = restart_neighbor
		self._delay.reset()

	def reconfigure (self, restart_neighbor=None):
		# we want to update the route which were in the configuration file
		self._reconfigure = True
		self._neighbor = restart_neighbor
		self._resend_routes = SEND.NORMAL
		self._neighbor = restart_neighbor

	def teardown (self, code, restart=True):
		self._restart = restart
		self._teardown = code
		self._delay.reset()

	# sockets we must monitor

	def sockets (self):
		if self.proto:
			fd = self.proto.fd()
			if fd:
				return [fd]
		return []

	def handle_connection (self, connection):
		self.logger.debug("state machine for the peer is %s" % self.fsm.name(), self.id())

		# if the other side fails, we go back to idle
		if self.fsm == FSM.ESTABLISHED:
			self.logger.debug('we already have a peer in state established for %s' % connection.name(),self.id())
			return connection.notification(6,7,'could not accept the connection, already established')

		# 6.8 The convention is to compare the BGP Identifiers of the peers
		# involved in the collision and to retain only the connection initiated
		# by the BGP speaker with the higher-valued BGP Identifier.
		# FSM.IDLE , FSM.ACTIVE , FSM.CONNECT , FSM.OPENSENT , FSM.OPENCONFIRM , FSM.ESTABLISHED

		if self.fsm == FSM.OPENCONFIRM:
			# We cheat: we are not really reading the OPEN, we use the data we have instead
			# it does not matter as the open message will be the same anyway
			local_id = self.neighbor.router_id.pack()
			remote_id = self.proto.negotiated.received_open.router_id.pack()

			if remote_id < local_id:
				self.logger.debug('closing incoming connection as we have an outgoing connection with higher router-id for %s' % connection.name(),self.id())
				return connection.notification(6,7,'could not accept the connection, as another connection is already in open-confirm and will go through')

		# accept the connection
		if self.proto:
			self.logger.debug('closing outgoing connection as we have another incoming on with higher router-id for %s' % connection.name(),self.id())
			self.proto.close('closing outgoing connection as we have another incoming on with higher router-id')

		self.proto = Protocol(self).accept(connection)
		self.generator = None
		# Let's make sure we do some work with this connection
		self._delay.reset()
		return None

	def established (self):
		return self.fsm == FSM.ESTABLISHED

	def negotiated_families(self):
		if self.proto:
			families = ["%s/%s" % (x[0], x[1]) for x in self.proto.negotiated.families]
		else:
			families = ["%s/%s" % (x[0], x[1]) for x in self.neighbor.families()]

		if len(families) > 1:
			return "[ %s ]" % " ".join(families)
		elif len(families) == 1:
			return families[0]

		return ''

	def _connect (self):
		proto = Protocol(self)
		generator = proto.connect()

		connected = False
		try:
			for connected in generator:
				if connected:
					break
				if self._teardown:
					raise Stop()
				# we want to come back as soon as possible
				yield ACTION.LATER
			self.proto = proto
		except Stop:
			# Connection failed
			if not connected and self.proto:
				self.proto.close('connection to %s:%d failed' % (self.neighbor.peer_address,self.neighbor.connect))

			# A connection arrived before we could establish !
			if not connected or self.proto:
				yield ACTION.NOW
				raise Interrupted()

	def _send_open (self):
		message = Message.CODE.NOP
		for message in self.proto.new_open():
			if ordinal(message.TYPE) == Message.CODE.NOP:
				yield ACTION.NOW
		yield message

	def _read_open (self):
		wait = environment.settings().bgp.openwait
		opentimer = ReceiveTimer(self.proto.connection.session,wait,1,1,'waited for open too long, we do not like stuck in active')
		# Only yield if we have not the open, otherwise the reactor can run the other connection
		# which would be bad as we need to do the collission check without going to the other peer
		for message in self.proto.read_open(self.neighbor.peer_address.top()):
			opentimer.check_ka(message)
			# XXX: FIXME: change the whole code to use the ord and not the chr version
			# Only yield if we have not the open, otherwise the reactor can run the other connection
			# which would be bad as we need to do the collission check
			if ordinal(message.TYPE) == Message.CODE.NOP:
				# If a peer does not reply to OPEN message, or not enough bytes
				# yielding ACTION.NOW can cause ExaBGP to busy spin trying to
				# read from peer. See GH #723 .
				yield ACTION.LATER
		yield message

	def _send_ka (self):
		for message in self.proto.new_keepalive('OPENCONFIRM'):
			yield ACTION.NOW

	def _read_ka (self):
		# Start keeping keepalive timer
		for message in self.proto.read_keepalive():
			self.recv_timer.check_ka_timer(message)
			yield ACTION.NOW

	def _establish (self):
		# try to establish the outgoing connection
		self.fsm.change(FSM.ACTIVE)

		if not self.proto:
			for action in self._connect():
				if action in ACTION.ALL:
					yield action
		self.fsm.change(FSM.CONNECT)

		# normal sending of OPEN first ...
		if self.neighbor.local_as:
			for sent_open in self._send_open():
				if sent_open in ACTION.ALL:
					yield sent_open
			self.proto.negotiated.sent(sent_open)
			self.fsm.change(FSM.OPENSENT)

		# read the peer's open
		for received_open in self._read_open():
			if received_open in ACTION.ALL:
				yield received_open
		self.proto.negotiated.received(received_open)

		self.proto.connection.msg_size = self.proto.negotiated.msg_size

		# if we mirror the ASN, we need to read first and send second
		if not self.neighbor.local_as:
			for sent_open in self._send_open():
				if sent_open in ACTION.ALL:
					yield sent_open
			self.proto.negotiated.sent(sent_open)
			self.fsm.change(FSM.OPENSENT)

		self.proto.validate_open()
		self.fsm.change(FSM.OPENCONFIRM)

		self.recv_timer = ReceiveTimer(self.proto.connection.session,self.proto.negotiated.holdtime,4,0)
		for action in self._send_ka():
			yield action
		for action in self._read_ka():
			yield action
		self.fsm.change(FSM.ESTABLISHED)
		self.stats['complete'] = time.time()

		# let the caller know that we were sucesfull
		yield ACTION.NOW

	def _main (self):
		"""yield True if we want to come back to it asap, None if nothing urgent, and False if stopped"""
		if self._teardown:
			raise Notify(6,3)

		self.neighbor.rib.incoming.clear()

		include_withdraw = False

		# Announce to the process BGP is up
		self.logger.notice('connected to %s with %s' % (self.id(),self.proto.connection.name()),'reactor')
		self.stats['up'] = self.stats.get('up',0) + 1
		if self.neighbor.api['neighbor-changes']:
			try:
				self.reactor.processes.up(self.neighbor)
			except ProcessError:
				# Can not find any better error code than 6,0 !
				# XXX: We can not restart the program so this will come back again and again - FIX
				# XXX: In the main loop we do exit on this kind of error
				raise Notify(6,0,'ExaBGP Internal error, sorry.')

		send_eor = not self.neighbor.manual_eor
		new_routes = None
		self._resend_routes = SEND.NORMAL
		send_families = []

		# Every last asm message should be re-announced on restart
		for family in self.neighbor.asm:
			if family in self.neighbor.families():
				self.neighbor.messages.appendleft(self.neighbor.asm[family])

		operational = None
		refresh = None
		command_eor = None
		number = 0
		refresh_enhanced = True if self.proto.negotiated.refresh == REFRESH.ENHANCED else False

		send_ka = KA(self.proto.connection.session,self.proto)

		while not self._teardown:
			for message in self.proto.read_message():
				self.recv_timer.check_ka(message)

				if send_ka() is not False:
					# we need and will send a keepalive
					while send_ka() is None:
						yield ACTION.NOW

				# Received update
				if message.TYPE == Update.TYPE:
					number += 1
					self.logger.debug('<< UPDATE #%d' % number,self.id())

					for nlri in message.nlris:
						self.neighbor.rib.incoming.update_cache(Change(nlri,message.attributes))
						self.logger.debug(LazyFormat('   UPDATE #%d nlri ' % number,nlri,str),self.id())

				elif message.TYPE == RouteRefresh.TYPE:
					if message.reserved == RouteRefresh.request:
						self._resend_routes = SEND.REFRESH
						send_families.append((message.afi,message.safi))

				# SEND OPERATIONAL
				if self.neighbor.operational:
					if not operational:
						new_operational = self.neighbor.messages.popleft() if self.neighbor.messages else None
						if new_operational:
							operational = self.proto.new_operational(new_operational,self.proto.negotiated)

					if operational:
						try:
							six.next(operational)
						except StopIteration:
							operational = None
				# make sure that if some operational message are received via the API
				# that we do not eat memory for nothing
				elif self.neighbor.messages:
					self.neighbor.messages.popleft()

				# SEND REFRESH
				if self.neighbor.route_refresh:
					if not refresh:
						new_refresh = self.neighbor.refresh.popleft() if self.neighbor.refresh else None
						if new_refresh:
							refresh = self.proto.new_refresh(new_refresh)

					if refresh:
						try:
							six.next(refresh)
						except StopIteration:
							refresh = None

				# Take the routes already sent to that peer and resend them
				if self._reconfigure:
					self._reconfigure = False

					# we are here following a configuration change
					if self._neighbor:
						# see what changed in the configuration
						self.neighbor.rib.outgoing.replace(self._neighbor.backup_changes,self._neighbor.changes)
						# do not keep the previous routes in memory as they are not useful anymore
						self._neighbor.backup_changes = []

				# Take the routes already sent to that peer and resend them
				if self._resend_routes != SEND.DONE:
					enhanced = True if refresh_enhanced and self._resend_routes == SEND.REFRESH else False
					self._resend_routes = SEND.DONE
					self.neighbor.rib.outgoing.resend(send_families,enhanced)
					send_families = []

				# Need to send update
				if not new_routes and self.neighbor.rib.outgoing.pending():
					# XXX: in proto really. hum to think about ?
					new_routes = self.proto.new_update(include_withdraw)

				if new_routes:
					try:
						for _ in range(25):
							# This can raise a NetworkError
							six.next(new_routes)
					except StopIteration:
						new_routes = None
						include_withdraw = True

				elif send_eor:
					send_eor = False
					for _ in self.proto.new_eors():
						yield ACTION.NOW
					self.logger.debug('>> EOR(s)',self.id())

				# SEND MANUAL KEEPALIVE (only if we have no more routes to send)
				elif not command_eor and self.neighbor.eor:
					new_eor = self.neighbor.eor.popleft()
					command_eor = self.proto.new_eors(new_eor.afi,new_eor.safi)

				if command_eor:
					try:
						six.next(command_eor)
					except StopIteration:
						command_eor = None

				if new_routes or message.TYPE != NOP.TYPE:
					yield ACTION.NOW
				elif self.neighbor.messages or operational:
					yield ACTION.NOW
				elif self.neighbor.eor or command_eor:
					yield ACTION.NOW
				else:
					yield ACTION.LATER

				# read_message will loop until new message arrives with NOP
				if self._teardown:
					break

		# If graceful restart, silent shutdown
		if self.neighbor.graceful_restart and self.proto.negotiated.sent_open.capabilities.announced(Capability.CODE.GRACEFUL_RESTART):
			self.logger.error('closing the session without notification',self.id())
			self.proto.close('graceful restarted negotiated, closing without sending any notification')
			raise NetworkError('closing')

		# notify our peer of the shutdown
		raise Notify(6,self._teardown)

	def _run (self):
		"""yield True if we want the reactor to give us back the hand with the same peer loop, None if we do not have any more work to do"""
		try:
			for action in self._establish():
				yield action

			for action in self._main():
				yield action

		# CONNECTION FAILURE
		except NetworkError as network:
			# we tried to connect once, it failed and it was not a manual request, we stop
			if self.once and not self._teardown:
				self.logger.debug('only one attempt to connect is allowed, stopping the peer',self.id())
				self.stop()

			self._reset('closing connection',network)
			return

		# NOTIFY THE PEER OF AN ERROR
		except Notify as notify:
			if self.proto:
				try:
					generator = self.proto.new_notification(notify)
					try:
						while True:
							six.next(generator)
							yield ACTION.NOW
					except StopIteration:
						pass
				except (NetworkError,ProcessError):
					self.logger.error('Notification not sent',self.id())
				self._reset('notification sent (%d,%d)' % (notify.code,notify.subcode),notify)
			else:
				self._reset()
			return

		# THE PEER NOTIFIED US OF AN ERROR
		except Notification as notification:
			# we tried to connect once, it failed and it was not a manual request, we stop
			if self.once and not self._teardown:
				self.logger.debug('only one attempt to connect is allowed, stopping the peer',self.id())
				self.stop()

			self._reset(
				'notification received (%d,%d)' % (
					notification.code,
					notification.subcode),
				notification
			)
			return

		# RECEIVED a Message TYPE we did not expect
		except Message as message:
			self._reset('unexpected message received',message)
			return

		# PROBLEM WRITING TO OUR FORKED PROCESSES
		except ProcessError as process:
			self._reset('process problem',process)
			return

		# ....
		except Interrupted as interruption:
			self._reset('connection received before we could fully establish one')
			return

		# UNHANDLED PROBLEMS
		except Exception as exc:
			# Those messages can not be filtered in purpose
			self.logger.debug('\n'.join([
				NO_PANIC,
				'',
				'',
				str(type(exc)),
				str(exc),
				trace(),
				FOOTER
			]),'reactor')
			self._reset()
			return
	# loop

	def run (self):
		if self.reactor.processes.broken(self.neighbor):
			# XXX: we should perhaps try to restart the process ??
			self.logger.error('ExaBGP lost the helper process for this peer - stopping','process')
			if self.reactor.processes.terminate_on_error:
				self.reactor.api_shutdown()
			else:
				self.stop()
			return True

		if self.generator:
			try:
				# This generator only stops when it raises
				# otherwise return one of the ACTION
				return six.next(self.generator)
			except StopIteration:
				# Trying to run a closed loop, no point continuing
				self.generator = None
				if self._restart:
					return ACTION.LATER
				return ACTION.CLOSE

		elif self.generator is None:
			if self.fsm in [FSM.OPENCONFIRM,FSM.ESTABLISHED]:
				self.logger.debug('stopping, other connection is established',self.id())
				self.generator = False
				return ACTION.LATER
			if self._delay.backoff():
				return ACTION.LATER
			if self._restart:
				self.logger.debug('initialising connection to %s' % self.id(),'reactor')
				self.generator = self._run()
				return ACTION.LATER  # make sure we go through a clean loop
			return ACTION.CLOSE

	def cli_data (self):
		def tri (value):
			if value is None:
				return None
			return True if value else False

		peer = defaultdict(lambda: None)

		have_peer = self.proto is not None
		have_open = self.proto and self.proto.negotiated.received_open

		if have_peer:
			peer.update({
				'multi-session': self.proto.negotiated.multisession,
				'operational':   self.proto.negotiated.operational,
			})

		if have_open:
			capa = self.proto.negotiated.received_open.capabilities
			peer.update({
				'router-id':     self.proto.negotiated.received_open.router_id,
				'hold-time':     self.proto.negotiated.received_open.hold_time,
				'asn4':          self.proto.negotiated.asn4,
				'route-refresh': capa.announced(Capability.CODE.ROUTE_REFRESH),
				'multi-session': capa.announced(Capability.CODE.MULTISESSION) or capa.announced(Capability.CODE.MULTISESSION_CISCO),
				'add-path':      capa.announced(Capability.CODE.ADD_PATH),
				'extended-message': capa.announced(Capability.CODE.EXTENDED_MESSAGE),
				'graceful-restart': capa.announced(Capability.CODE.GRACEFUL_RESTART),
			})

		capabilities = {
			'asn4':             (tri(self.neighbor.asn4), tri(peer['asn4'])),
			'route-refresh':    (tri(self.neighbor.route_refresh),tri(peer['route-refresh'])),
			'multi-session':    (tri(self.neighbor.multisession), tri(peer['multi-session'])),
			'operational':      (tri(self.neighbor.operational), tri(peer['operational'])),
			'add-path':         (tri(self.neighbor.add_path),tri(peer['add-path'])),
			'extended-message': (tri(self.neighbor.extended_message),tri(peer['extended-message'])),
			'graceful-restart': (tri(self.neighbor.graceful_restart),tri(peer['graceful-restart'])),
		}

		families = {}
		for family in self.neighbor.families():
			if have_open:
				common = True if family in self.proto.negotiated.families else False
				addpath = self.proto.negotiated.addpath.receive(*family) and self.proto.negotiated.addpath.receive(*family)
			else:
				common = None
				addpath = None if family in self.neighbor.addpaths() else False
			families[family] = (True,common,addpath)

		messages = {}
		total_sent = 0
		total_rcvd = 0
		for message in ('open','notification','keepalive','update','refresh'):
			sent = self.stats.get('send-%s' % message,0)
			rcvd = self.stats.get('receive-%s' % message,0)
			total_sent += sent
			total_rcvd += rcvd
			messages[message] = (sent, rcvd)
		messages['total'] = (total_sent, total_rcvd)

		return {
			'down':          int(self.stats['reset'] - self.stats['creation']),
			'duration':      int(time.time() - self.stats['complete']) if self.stats['complete'] else 0,
			'local-address': str(self.neighbor.local_address),
			'peer-address':  str(self.neighbor.peer_address),
			'local-as':      int(self.neighbor.local_as),
			'peer-as':       int(self.neighbor.peer_as),
			'local-id':      str(self.neighbor.router_id),
			'peer-id':       None if peer['peer-id'] is None else str(peer['router-id']),
			'local-hold':    int(self.neighbor.hold_time),
			'peer-hold':     None if peer['hold-time'] is None else int(peer['hold-time']),
			'state':         self.fsm.name(),
			'capabilities':  capabilities,
			'families':      families,
			'messages':      messages,
		}



# encoding: utf-8
"""
bgp.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.util import character
from exabgp.util import concat_bytes_i
from exabgp.util import hexstring


# =========================================================== RouteDistinguisher
# RFC 4364

class RouteDistinguisher (object):

	__slots__ = ['rd','_len']

	def __init__ (self, rd):
		self.rd = rd
		self._len = len(self.rd)

	def __eq__ (self, other):
		return self.rd == other.rd

	def __neq__ (self, other):
		return self.rd != other.rd

	def __lt__ (self, other):
		raise RuntimeError('comparing RouteDistinguisher for ordering does not make sense')

	def __le__ (self, other):
		raise RuntimeError('comparing RouteDistinguisher for ordering does not make sense')

	def __gt__ (self, other):
		raise RuntimeError('comparing RouteDistinguisher for ordering does not make sense')

	def __ge__ (self, other):
		raise RuntimeError('comparing RouteDistinguisher for ordering does not make sense')

	def pack (self):
		return self.rd

	def __len__ (self):
		return self._len

	def _str (self):
		t,c1,c2,c3 = unpack('!HHHH',self.rd)
		if t == 0:
			rd = '%d:%d' % (c1,(c2 << 16)+c3)
		elif t == 1:
			rd = '%d.%d.%d.%d:%d' % (c1 >> 8,c1 & 0xFF,c2 >> 8,c2 & 0xFF,c3)
		elif t == 2:
			rd = '%d:%d' % ((c1 << 16) + c2,c3)
		else:
			rd = hexstring(self.rd)
		return rd

	def json (self):
		if not self.rd:
			return ''
		return '"rd": "%s"' % self._str()

	def __hash__(self):
		return hash(self.rd)

	def __repr__ (self):
		if not self.rd:
			return ''
		return ' rd %s' % self._str()

	@classmethod
	def unpack (cls, data):
		return cls(data[:8])

	# DO NOT USE, the right function is route_distinguisher() in exabgp.configuation.static.mpls
	@classmethod
	def fromElements (cls, prefix, suffix):
		try:
			if '.' in prefix:
				data = [character(0),character(1)]
				data.extend([character(int(_)) for _ in prefix.split('.')])
				data.extend([character(suffix >> 8),character(suffix & 0xFF)])
				distinguisher = concat_bytes_i(data)
			else:
				number = int(prefix)
				if number < pow(2,16) and suffix < pow(2,32):
					distinguisher = character(0) + character(0) + pack('!H',number) + pack('!L',suffix)
				elif number < pow(2,32) and suffix < pow(2,16):
					distinguisher = character(0) + character(2) + pack('!L',number) + pack('!H',suffix)
				else:
					raise ValueError('invalid route-distinguisher %s' % number)

			return cls(distinguisher)
		except ValueError:
			raise ValueError('invalid route-distinguisher %s:%s' % (prefix,suffix))


RouteDistinguisher.NORD = RouteDistinguisher(b'')



# encoding: utf-8
"""
prefixmetric.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.vendoring.bitstring import BitArray
from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                            Metric                             |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     https://tools.ietf.org/html/rfc7752#section-3.3.3.4


@LINKSTATE.register()
class PrefixMetric(object):
	TLV = 1155

	def __init__ (self, prefixmetric):
		self.prefixmetric = prefixmetric

	def __repr__ (self):
		return "Prefix metric: %s" % (self.prefixmetric)

	@classmethod
	def unpack (cls,data,length):
		if length != 4:
			raise Notify(3,5, "Incorrect Prefix Metric size")
		else:
			metric = unpack("!L", data)[0]
			return cls(prefixmetric=metric)

	def json (self,compact=None):
		return '"prefix-metric": %d' % int(self.prefixmetric)



# encoding: utf-8
"""
resource.py

Created by Thomas Mangin on 2015-05-15.
Copyright (c) 2015-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import sys

from exabgp.util import string_is_hex


if sys.version_info > (3,):
	long = int


class Resource (long):
	NAME = ''
	codes = {}
	names = {}

	cache = {}

	def __new__ (cls, *args):
		key = '//'.join((str(_) for _ in args))
		if key in Resource.cache.setdefault(cls,{}):
			return Resource.cache[cls][key]
		instance = long.__new__(cls,*args)
		Resource.cache[cls][key] = instance
		return instance

	def __str__ (self):
		return self.names.get(self,'unknown %s type %ld' % (self.NAME,long(self)))

	@classmethod
	def _value (cls,string):
		name = string.lower().replace('_','-')
		if name in cls.codes:
			return cls.codes[name]
		if string.isdigit():
			value = int(string)
			if 0 <= value <= 0xFFFF:
				return value
		if string_is_hex(string):
			value = int(string[2:],16)
			if 0 <= value <= 0xFFFF:
				return value
		raise ValueError('unknown %s %s' % (cls.NAME,name))

	@classmethod
	def named (cls,string):
		value = 0
		for name in string.split('+'):
			value += cls._value(name)
		return cls(value)


class BitResource (Resource):
	def bits (self):
		value = int(self)
		for bit in self.names.keys():
			if value & bit:
				yield self.names[bit]
				value -= bit
		if value:
			yield self.names.get(self,'unknown %s type %ld' % (self.NAME,long(self)))

	def named_bits (self):
		for value in self.bits():
			yield value

	def __str__ (self):
		return '+'.join(self.bits())



# encoding: utf-8
"""
flowspec_scope.py

Created by Stephane Litkowski on 2017-02-24.
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity


# ============================================================== InterfaceSet
# draft-ietf-idr-flowspsec-interfaceset


@ExtendedCommunity.register
class InterfaceSet(ExtendedCommunity):
	COMMUNITY_TYPE = 0x07
	COMMUNITY_SUBTYPE = 0x02

	__slots__ = ['trans', 'asn','target','direction']

	names = {
		1:	'input',
		2:	'output',
		3:	'input-output',
	}

	def __init__ (self, trans, asn, target, direction, community=None):
		self.asn = asn
		self.target = target
		self.direction = direction
		self.transitive = trans
		new_target = (direction << 14) + target
		ExtendedCommunity.__init__(
			self,
			community if community is not None else pack(
				"!2sLH",
				self._subtype(self.transitive),
				asn,new_target
			)
		)

	def __repr__ (self):
		str_direction = self.names.get(self.direction, str(self.direction))
		return "interface-set:%s:%s:%s" % (
			str_direction,
			str(self.asn),
			str(self.target)
		)

	@staticmethod
	def unpack (data):
		asn,target = unpack('!LH',data[2:8])
		direction = target >> 14
		target = target & 0x1FFF
		return InterfaceSet(False,ASN(asn),target,direction,data[:8])



# encoding: utf-8
"""
line/rib.py

Created by Thomas Mangin on 2017-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.reactor.api.command.command import Command
from exabgp.reactor.api.command.limit import match_neighbors
from exabgp.reactor.api.command.limit import extract_neighbors

from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.bgp.message.update.nlri.inet import INET
from exabgp.bgp.message.update.nlri.flow import Flow
from exabgp.bgp.message.update.nlri.vpls import VPLS
from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN

from exabgp.reactor.api.response.answer import Answer

from exabgp.configuration.environment import environment


def register_rib ():
	pass


def _show_adjrib_callback(reactor, service, last, route_type, advertised, rib_name, extensive):
	def callback ():
		families = None
		lines_per_yield = environment.settings().api.chunk
		if last in ('routes', 'extensive', 'static', 'flow', 'l2vpn'):
			peers = list(reactor.peers)
		else:
			peers = [n for n in reactor.peers.keys() if 'neighbor %s' % last in n]
		for key in peers:
			peer = reactor.peers.get(key, None)
			if not peer:
				continue
			if advertised:
				families = peer.proto.negotiated.families if peer.proto else []
			rib = peer.neighbor.rib.outgoing if rib_name == 'out' else peer.neighbor.rib.incoming
			routes = list(rib.cached_changes(families))
			while routes:
				changes, routes = routes[:lines_per_yield], routes[lines_per_yield:]
				for change in changes:
					if isinstance(change.nlri, route_type):
						if extensive:
							reactor.processes.answer(service,'%s %s %s' % (peer.neighbor.name(),'%s %s' % change.nlri.family(),change.extensive()),force=True)
						else:
							reactor.processes.answer(service,'neighbor %s %s %s' % (peer.neighbor.peer_address,'%s %s' % change.nlri.family(),str(change.nlri)),force=True)
				yield True
		reactor.processes.answer_done(service)
	return callback


@Command.register('text','show adj-rib')
def show_adj_rib (self, reactor, service, line):
	words = line.split()
	extensive = line.endswith(' extensive')
	try:
		rib = words[2]
	except IndexError:
		if words[1] == 'adj-rib-in':
			rib = 'in'
		elif words[1] == 'adj-rib-out':
			rib = 'out'
		else:
			reactor.processes.answer(service,Answer.error)
			return False

	if rib not in ('in','out'):
		reactor.processes.answer(service,Answer.error)
		return False

	klass = NLRI

	if 'inet' in words:
		klass = INET
	elif 'flow' in words:
		klass = Flow
	elif 'l2vpn' in words:
		klass = (VPLS, EVPN)

	for remove in ('show','adj-rib','adj-rib-in','adj-rib-out','in','out','extensive'):
		if remove in words:
			words.remove(remove)
	last = '' if not words else words[0]
	callback = _show_adjrib_callback(reactor, service, last, klass, False, rib, extensive)
	reactor.asynchronous.schedule(service, line, callback())
	return True


@Command.register('text','flush adj-rib out')
def flush_adj_rib_out (self, reactor, service, line):
	def callback (self, peers):
		self.log_message("Flushing adjb-rib out for %s" % ', '.join(peers if peers else []) if peers is not None else 'all peers')
		for peer_name in peers:
			peer = reactor.peers.get(peer_name, None)
			if not peer:
				continue
			yield False

		reactor.processes.answer_done(service)

	try:
		descriptions,command = extract_neighbors(line)
		peers = match_neighbors(reactor.peers,descriptions)
		if not peers:
			self.log_failure('no neighbor matching the command : %s' % command,'warning')
			reactor.processes.answer(service,Answer.error)
			return False
		reactor.asynchronous.schedule(service, command, callback(self, peers))
		return True
	except ValueError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service, Answer.error)
		return False
	except IndexError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service, Answer.error)
		return False


@Command.register('text','clear adj-rib')
def clear_adj_rib (self, reactor, service, line):
	def callback (self, peers, direction):
		self.log_message("clearing adjb-rib-%s for %s" % (direction,', '.join(peers if peers else []) if peers is not None else 'all peers'))
		for peer_name in peers:
			peer = reactor.peers.get(peer_name, None)
			if not peer:
				continue
			if direction == 'out':
				peer.neighbor.rib.outgoing.clear()
			else:
				peer.neighbor.rib.incoming.clear()
			yield False

		reactor.processes.answer_done(service)

	try:
		descriptions,command = extract_neighbors(line)
		peers = match_neighbors(reactor.peers,descriptions)
		if not peers:
			self.log_failure('no neighbor matching the command : %s' % command,'warning')
			reactor.processes.answer(service, Answer.error)
			return False
		words = line.split()
		direction = 'in' if 'adj-rib-in' in words or 'in' in words else 'out'
		reactor.asynchronous.schedule(service, command, callback(self, peers, direction))
		return True
	except ValueError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service, Answer.error)
		return False
	except IndexError:
		self.log_failure('issue parsing the command')
		reactor.processes.answer(service, Answer.error)
		return False



#!/usr/bin/env python
# encoding: utf-8
"""
tojson.py

Created by Thomas Mangin on 2014-12-22.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import print_function

import os
import sys
import time
import signal
import thread
import subprocess
from collections import deque

from exabgp.reactor.api.transcoder import Transcoder
from exabgp.reactor.api.processes import preexec_helper

from exabgp.configuration.setup import environment

environment.setup('')

# test = """\
# { "exabgp": "3.5.0", "time": 1430238962.74, "host" : "mangin.local", "pid" : 37912, "ppid" : 37903, "type": "state", "neighbor": { "address": { "local": "82.219.212.34", "peer": "127.0.0.1" }, "asn": { "local": "65534", "peer": "65533" }, "state": "down", "reason": "out loop, peer reset, message [closing connection] error[the TCP connection was closed by the remote end]"} }
# { "exabgp": "3.5.0", "time": 1430238928.75, "host" : "mangin.local", "pid" : 37912, "ppid" : 37903, "type": "state", "neighbor": { "address": { "local": "82.219.212.34", "peer": "127.0.0.1" }, "asn": { "local": "65534", "peer": "65533" }, "state": "up"} }
# { "exabgp": "3.5.0", "time": 1430293452.31, "host" : "mangin.local", "pid" : 57788, "ppid" : 57779, "type": "open", "neighbor": { "address": { "local": "172.20.10.6", "peer": "127.0.0.1" }, "asn": { "local": "65534", "peer": "65533" }, "direction": "send", "message": { "category": 1, "header": "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00AF01", "body": "04FFFE00B4800000009202060104000100010206010400010002020601040001000402060104000100800206010400010085020601040001008602060104000200010206010400020080020601040002008502060104000200860206010400190041020641040000FFFE0230402E84B00001018000010280000104800001808000018580000186800002018000028080000285800002868000194180" } } }
# { "exabgp": "3.5.0", "time": 1430293452.32, "host" : "mangin.local", "pid" : 57788, "ppid" : 57779, "type": "open", "neighbor": { "address": { "local": "172.20.10.6", "peer": "127.0.0.1" }, "asn": { "local": "65534", "peer": "65533" }, "direction": "receive", "message": { "category": 1, "header": "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00AF01", "body": "04FFFD00B47F0000009202060104000100010206010400010002020601040001000402060104000100800206010400010085020601040001008602060104000200010206010400020080020601040002008502060104000200860206010400190041020641040000FFFD0230402E84B00001018000010280000104800001808000018580000186800002018000028080000285800002868000194180" } } }
# { "exabgp": "3.5.0", "time": 1430238928.75, "host" : "mangin.local", "pid" : 37912, "ppid" : 37903, "type": "update", "neighbor": { "address": { "local": "82.219.212.34", "peer": "127.0.0.1" }, "asn": { "local": "65534", "peer": "65533" }, "direction": "receive", "message": { "category": 2, "header": "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF005002", "body": "0000002740010100400212020400000001000000020000000300000004400304010101018004040000006408630858084D08420837082C08210816080B" } } }
# { "exabgp": "3.5.0", "time": 1430238928.76, "host" : "mangin.local", "pid" : 37912, "ppid" : 37903, "type": "update", "neighbor": { "address": { "local": "82.219.212.34", "peer": "127.0.0.1" }, "asn": { "local": "65534", "peer": "65533" }, "direction": "receive", "message": { "category": 2, "header": "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF001E02", "body": "00000007900F0003001941" } } }
# { "exabgp": "3.5.0", "time": 1430238928.76, "host" : "mangin.local", "pid" : 37912, "ppid" : 37903, "type": "update", "neighbor": { "address": { "local": "82.219.212.34", "peer": "127.0.0.1" }, "asn": { "local": "65534", "peer": "65533" }, "direction": "receive", "message": { "category": 2, "header": "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF001E02", "body": "00000007900F0003000286" } } }
# { "exabgp": "3.5.0", "time": 1430238928.76, "host" : "mangin.local", "pid" : 37912, "ppid" : 37903, "type": "update", "neighbor": { "address": { "local": "82.219.212.34", "peer": "127.0.0.1" }, "asn": { "local": "65534", "peer": "65533" }, "direction": "receive", "message": { "category": 2, "header": "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF001E02", "body": "00000007900F0003000285" } } }
# """.split('\n')


# XXX: We do not accept input from the forked application

class Application (object):
	Q = deque()
	running = True

	@staticmethod
	def _signal (_,__):
		Application.running = False

	def process (self):
		run = sys.argv[1:]
		if not run:
			print(sys.stderr, 'no consummer program provided')
			sys.exit(1)

		# Prevent some weird termcap data to be created at the start of the PIPE
		# \x1b[?1034h (no-eol) (esc)
		os.environ['TERM'] = 'dumb'

		try:
			sub = subprocess.Popen(
				run,
				stdin=subprocess.PIPE,
				stdout=subprocess.PIPE,
				preexec_fn=preexec_helper
				# This flags exists for python 2.7.3 in the documentation but on on my MAC
				# creationflags=subprocess.CREATE_NEW_PROCESS_GROUP
			)
		except (subprocess.CalledProcessError,OSError,ValueError):
			print('could not start subprocess', file=sys.stderr)
			sys.exit(1)

		return sub

	def __init__ (self):
		thread.start_new_thread(self.reader,(os.getpid(),))
		signal.signal(signal.SIGTERM,Application._signal)
		self.sub = self.process()
		self.transcoder = Transcoder('json','json')
		self.main()

	# def test_reader (self,myself):
	# 	while len(test):
	# 		# line = sys.stdin.readline().strip()
	# 		line = test.pop(0)
	# 		if line:
	# 			self.Q.append(line)
	# 	time.sleep(2)
	# 	os.kill(myself,signal.SIGTERM)

	def reader (self,myself):
		ok = True
		line = ''
		while True:
			line = sys.stdin.readline().strip()
			if ok:
				if not line:
					ok = False
					continue
			elif not line:
				break
			else:
				ok = True
			self.Q.append(line)
		os.kill(myself,signal.SIGTERM)

	def main (self):
		while self.running or len(self.Q):
			try:
				line = self.Q.popleft()
				self.sub.stdin.write(self.transcoder.convert(line))
				self.sub.stdin.flush()
			except IndexError:
				# no data on the Q to read
				time.sleep(0.1)
			except IOError:
				# subprocess died
				print('subprocess died', file=sys.stderr)
				sys.exit(1)


if __name__ == '__main__':
	try:
		Application()
	except KeyboardInterrupt:
		pass



from struct import pack
from struct import unpack

from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity

# draft-fm-bess-service-chaining


@ExtendedCommunity.register
class ConsistentHashSortOrder(ExtendedCommunity):
	COMMUNITY_TYPE = 0x03
	COMMUNITY_SUBTYPE = 0x14
	DESCRIPTION = "consistentHashSortOrder"

	__slots__ = ['order','reserved']

	def __init__ (self, order, reserved=0, community=None):
		self.order = order
		self.reserved = reserved

		ExtendedCommunity.__init__(
			self,
			community if community is not None else pack(
				"!2sIH",
				self._subtype(),
				order,reserved
			)
		)

	def __repr__ (self):
		return "%s:%d" % (self.DESCRIPTION, self.order)

	@staticmethod
	def unpack (data):
		order,reserved = unpack('!IH',data[2:8])
		return ConsistentHashSortOrder(order, reserved, data[:8])



import sys
import pdb

from exabgp.configuration.environment import environment

if sys.version_info[0] >= 3:
	StandardError = Exception


class Error (StandardError):
	def __init__ (self):
		self.message = ''
		self.debug = environment.settings().debug.configuration

	def set (self, message):
		self.message = message
		if self.debug:
			error = False
			print('\n%s\n' % self.message)
			pdb.set_trace()
			return error
		return False

	def throw (self,message):
		self.message = message
		if self.debug:
			print('\n%s\n' % message)
			pdb.set_trace()
		else:
			raise self

	def clear (self):
		self.message = ''

	def __repr__ (self):
		return self.message

	def __str__ (self):
		return self.message



#!/usr/bin/env python

import os
import sys
import time


def _prefixed (level, message):
	now = time.strftime('%a, %d %b %Y %H:%M:%S',time.localtime())
	return "%s %-8s %-6d %s" % (now,level,os.getpid(),message)


# When the parent dies we are seeing continual newlines, so we only access so many before stopping
counter = 0


while True:
	try:
		line = sys.stdin.readline().strip()
		# print >> sys.stderr, "GOT A LINE"
		sys.stdout.flush()
		if line == "":
			counter += 1
			if counter > 100:
				break
			continue

		counter = 0

		print >> sys.stderr, _prefixed(sys.argv[1] if len(sys.argv) >= 2 else 'EXABGP PROCESS',line)
	except KeyboardInterrupt:
		pass
	except IOError:
		# most likely a signal during readline
		pass
